import os
from collections import OrderedDict
from typing import List, Tuple, Union

import astropy.units as u
import h5py
import numpy as np

import exosim.log as log
import exosim.tasks.detector as detector
from exosim.models.signal import Counts
from exosim.output import HDF5OutputGroup, SetOutput
from exosim.output.hdf5.utils import copy_file, recursively_read_dict_contents
from exosim.utils.ascii_arts import astronomer3
from exosim.utils.checks import look_for_key
from exosim.utils.klass_factory import find_and_run_task
from exosim.utils.prepare_recipes import (
    clean_config_files,
    copy_input_files,
    load_options,
)
from exosim.utils.runConfig import RunConfig
from exosim.utils.timed_class import TimedClass


class CreateNDRs(TimedClass, log.Logger):
    """
    Pipeline to create the observation NDRs.
    This pipeline loads the configuration file and the Sub-Exposures.

    Examples
    --------

    >>> import exosim.recipes as recipes
    >>> recipes.CreateNDRs(input_file='./input_file.h5',
    >>>                    output_file = 'ndrs_file.h5',
    >>>                    options_file='main _configuration.xml')

    """

    def __init__(
        self, input_file: str, output_file: str, options_file: Union[str, dict]
    ) -> None:
        """
        Parameters
        ----------
        input_file : str
            Sub-Exposure ExoSim product
        output_file : str
            output file name
        options_file: str or dict
            input configuration file
        """

        super().__init__()
        self.graphics(astronomer3)
        RunConfig.stats()

        self.announce("started")
        clean_config_files()
        self.input = input_file
        self.mainConfig, self.payloadConfig = load_options(options_file)
        copy_input_files(os.path.dirname(os.path.abspath(output_file)))

        output = SetOutput(output_file)

        with h5py.File(self.input, "r") as f:
            ch_list = list(f["channels"].keys())
            ch_list.sort()
        with output.use(append=True, cache=True) as out:
            output_grp = out.create_group("channels")
            for ch in ch_list:
                ch_description = (
                    self.payloadConfig["channel"][ch]
                    if isinstance(self.payloadConfig["channel"], OrderedDict)
                    else self.payloadConfig["channel"]
                )
                # prepare output
                self.announce("iterating over {}".format(ch))
                ch_grp = output_grp.create_group(ch)

                # import exposures data
                (
                    spectral,
                    spatial,
                    mid_freq_time_line,
                    integration_times,
                    number_of_exposures,
                    n_subexposures_per_groups,
                    n_groups_per_ramp,
                ) = self.load_subexposure_data(ch)

                # preparing outputs
                sub_ndrs, sim_grp = self.prepare_output(
                    spectral,
                    spatial,
                    mid_freq_time_line,
                    integration_times,
                    ch,
                    ch_grp,
                )

                # add dark current
                if look_for_key(
                    ch_description["detector"], "dark_current", True
                ):
                    dc_instance = find_and_run_task(
                        ch_description["detector"],
                        "dc_task",
                        detector.AddConstantDarkCurrent,
                    )
                    dc_instance(
                        subexposures=sub_ndrs,
                        parameters=ch_description,
                        integration_times=integration_times,
                        output=sim_grp,
                    )

                # add shot noise
                if look_for_key(
                    ch_description["detector"], "shot_noise", True
                ):
                    addShotNoise = detector.AddShotNoise()
                    addShotNoise(subexposures=sub_ndrs, output=sim_grp)

                # add cosmic rays
                if look_for_key(
                    ch_description["detector"], "cosmic_rays", True
                ):
                    cosmic_instance = find_and_run_task(
                        ch_description["detector"],
                        "cosmic_rays_task",
                        detector.AddCosmicRays,
                    )
                    cosmic_instance(
                        subexposures=sub_ndrs,
                        parameters=ch_description,
                        integration_times=integration_times,
                        output=sim_grp,
                    )

                # prepare accumulation
                n_subexposures_per_ramp = (
                    sub_ndrs.dataset.shape[0] // number_of_exposures
                )
                base = np.arange(0, number_of_exposures, 1)
                state_machine = np.repeat(base, n_subexposures_per_ramp)

                # accumulate ndrs
                accumulateSubExposures = detector.AccumulateSubExposures()
                accumulateSubExposures(
                    subexposures=sub_ndrs, state_machine=state_machine
                )

                # add reset bias
                if look_for_key(
                    ch_description["detector"], "ktc_offset", True
                ):
                    reset_instance = find_and_run_task(
                        ch_description["detector"],
                        "ktc_offset_task",
                        detector.AddKTC,
                    )
                    reset_instance(
                        subexposures=sub_ndrs,
                        parameters=ch_description,
                        state_machine=state_machine,
                        output=sim_grp,
                    )

                # apply dead pixels map
                if look_for_key(
                    ch_description["detector"], "dead_pixels", True
                ):
                    dp_map_instance = find_and_run_task(
                        ch_description["detector"],
                        "dp_map_task",
                        detector.ApplyDeadPixelsMap,
                    )
                    dp_map_instance(
                        subexposures=sub_ndrs,
                        parameters=ch_description,
                        output=sim_grp,
                    )

                # apply pixel non linearity
                if look_for_key(
                    ch_description["detector"], "pixel_non_linearity", True
                ):
                    pnl_map_instance = find_and_run_task(
                        ch_description["detector"],
                        "pnl_map_task",
                        detector.LoadPixelsNonLinearityMap,
                    )
                    pnl_map = pnl_map_instance(
                        parameters=ch_description,
                        output=sim_grp,
                    )
                    pnl_apply_instance = find_and_run_task(
                        ch_description["detector"],
                        "pnl_task",
                        detector.ApplyPixelsNonLinearity,
                    )
                    pnl_apply_instance(
                        subexposures=sub_ndrs, parameters=pnl_map
                    )

                # apply pixel saturation
                if look_for_key(
                    ch_description["detector"], "saturation", True
                ):
                    sat_instance = find_and_run_task(
                        ch_description["detector"],
                        "sat_task",
                        detector.ApplySimpleSaturation,
                    )
                    sat_instance(
                        subexposures=sub_ndrs, parameters=ch_description
                    )

                # add read noise
                if look_for_key(
                    ch_description["detector"], "read_noise", True
                ):
                    read_instance = find_and_run_task(
                        ch_description["detector"],
                        "read_noise_task",
                        detector.AddNormalReadNoise,
                    )
                    read_instance(
                        subexposures=sub_ndrs,
                        parameters=ch_description,
                        output=sim_grp,
                    )

                # add gain drift
                if look_for_key(
                    ch_description["detector"], "gain_drift", True
                ):
                    gain_instance = find_and_run_task(
                        ch_description["detector"],
                        "gain_drift_task",
                        detector.addGainDrift,
                    )
                    gain_instance(
                        subexposures=sub_ndrs,
                        parameters=ch_description,
                        output=sim_grp,
                    )

                # converting to digital
                if look_for_key(ch_description["detector"], "ADC", True):
                    analogToDigital = detector.AnalogToDigital()
                    sub_ndrs = analogToDigital(
                        subexposures=sub_ndrs,
                        parameters=ch_description,
                        output=sim_grp,
                    )

                # merging groups' ndrs
                mergeGroups = detector.MergeGroups()
                merged_ndrs = mergeGroups(
                    subexposures=sub_ndrs,
                    n_groups=n_groups_per_ramp * number_of_exposures,
                    n_ndrs=n_subexposures_per_groups,
                    output=ch_grp,
                )

                merged_ndrs.metadata.update(
                    {"n_groups_per_ramp": n_groups_per_ramp}
                )
                merged_ndrs.write()

                # clean the output from redundant information
                self.clean_output_tree(ch_grp, ["simulation_data/sub_ndrs"])

                self.info("{} completed.".format(ch))
                self.log_runtime("{} ended in".format(ch), "info")

        self.refactor_output(output_file)
        self.info("output {} size: {:.3f}".format(output_file, out.getsize()))
        self.log_runtime_complete("recipe ended", "info")
        self.announce("ended")

    def load_subexposure_data(
        self, ch: str
    ) -> Tuple[u.Quantity, u.Quantity, u.Quantity, u.Quantity, int, int, int]:
        """It loads the sub-exposures data from file

        Parameters
        ----------
        ch : str
            channel name

        Returns
        -------
        :class:`astropy.units.Quantity`
            spectral axes array
        :class:`astropy.units.Quantity`
            spatial axes array
        :class:`astropy.units.Quantity`
            temporal array
        :class:`astropy.units.Quantity`
            integration times
        int
            number of exposures
        int
            number of NDRs per ramp
        int
            number of NDRs per group
        """
        with h5py.File(self.input, "r") as f:
            file_path = os.path.join("channels", ch)
            se = f[os.path.join(file_path, "SubExposures")]
            spectral = se["spectral"] * u.Unit(
                se["spectral_units"][()].decode("utf-8")
            )
            spatial = se["spatial"] * u.Unit(
                se["spatial_units"][()].decode("utf-8")
            )
            time_line = se["time"] * u.Unit(
                se["time_units"][()].decode("utf-8")
            )
            integration_times = se["metadata"]["integration_times"]["value"][
                ()
            ] * u.Unit(
                se["metadata"]["integration_times"]["unit"][()].decode("utf-8")
            )
            params = dict(
                f[os.path.join(file_path, "instantaneous_readout_params")]
            )
            try:
                number_of_exposures = params["number_of_exposures"][()]
            except TypeError:
                number_of_exposures = params["number_of_exposures"]["value"][
                    ()
                ]

            params = dict(f[os.path.join(file_path, "reading_scheme_params")])
            n_subexposures_per_groups = params["n_NRDs_per_group"][()]
            n_groups_per_ramp = params["n_GRPs"][()]

        return (
            spectral,
            spatial,
            time_line,
            integration_times,
            number_of_exposures,
            n_subexposures_per_groups,
            n_groups_per_ramp,
        )

    def prepare_output(
        self,
        spectral: u.Quantity,
        spatial: u.Quantity,
        time_line: u.Quantity,
        integration_times: u.Quantity,
        ch: str,
        ch_grp: HDF5OutputGroup,
    ) -> Tuple[Counts, HDF5OutputGroup]:
        """It produces the NDRs output per channel.

        Parameters
        ----------
        spectral : u.Quantity
            spectral axes array
        spatial : u.Quantity
            spatial axes array
        time_line : u.Quantity
            temporal array
        integration_times : u.Quantity
            integration times
        ch : str
            channel name
        ch_grp : h5py.Group
            output group

        Returns
        -------
        :class:`exosim.models.signal.Counts`
            NDRs signal class
        :class:`exosim.output.hdf5.h5df.HDF5OutputGroup`
            output group
        """

        sim_grp = ch_grp.create_group("simulation_data")

        with h5py.File(self.input, "r") as f:
            file_path = os.path.join("channels", ch)

            # copy simulation data
            sim_data = os.path.join(file_path, "simulation_data")
            for key in f[sim_data].keys():
                f.copy(os.path.join(sim_data, key), sim_grp._entry)
            f.copy("info", sim_grp._entry, name="sub_exposures_info")

            inst_rdo_data = os.path.join(
                file_path, "instantaneous_readout_params"
            )
            keys_to_keep = [
                "spe_jit_averaged",
                "spa_jit_averaged",
                "y_jit",
                "x_jit",
                "jit_indexes",
                "y_jit_averaged",
                "x_jit_averaged",
                "effective_osf",
            ]
            for key in f[inst_rdo_data].keys():
                if key in keys_to_keep:
                    f.copy(os.path.join(inst_rdo_data, key), sim_grp._entry)
            if "pointing_jitter" in f.keys():
                for key in f["pointing_jitter"]:
                    f.copy(
                        os.path.join("pointing_jitter", key), sim_grp._entry
                    )

            self.debug("simulation data copied to output")

            se = f[os.path.join(file_path, "SubExposures")]
            subexposures = se["data"]

            metadata = recursively_read_dict_contents(se["metadata"])
            # create sub_ndrs data
            sub_ndrs = Counts(
                spectral=spectral,
                time=time_line,
                data=None,
                spatial=spatial,
                shape=(
                    subexposures.shape[0],
                    subexposures.shape[1],
                    subexposures.shape[2],
                ),
                cached=True,
                output=sim_grp,
                dataset_name="sub_ndrs",
                output_path=None,
                metadata={"integration_times": integration_times},
                dtype=np.float64,
            )
            sub_ndrs.metadata.update(metadata)

            for chunk in sub_ndrs.dataset.iter_chunks():
                sub_ndrs.dataset[chunk] = subexposures[chunk]
                sub_ndrs.output.flush()
            sub_ndrs.write()

        return sub_ndrs, sim_grp

    def clean_output_tree(
        self, out: HDF5OutputGroup, key_list: List[str]
    ) -> None:
        """
        Remove specified keys and their associated data from the output object.

        This method iterates through a list of keys and deletes each key-value pair from the output object.
        Useful for cleaning up temporary or unnecessary data from the output.

        Parameters
        ----------
        out : HDF5OutputGroup
            The output object where key-value pairs are stored.
        key_list : list of str
            A list of keys that need to be removed from the output object.

        Examples
        --------
        >>> out = Output({'key1': 'value1', 'key2': 'value2', 'key3': 'value3'})
        >>> clean_output(out, ['key1', 'key3'])
        >>> out.data
        {'key2': 'value2'}

        Notes
        -----
        The method assumes that the keys specified in the `key_list` are present in the output object.
        If a key is not found, the `delete_data` method should handle it gracefully.
        """
        self.info("Cleaning the output tree")
        for key in key_list:
            out.delete_data(key)

    def refactor_output(self, output_file: str) -> None:
        """
        Renames the original output file and creates a new one with its contents.

        This method takes the original output file, renames it by appending '_unrefactored' to its name,
        and then copies its content into a new file with the original name. The refactored (old version)
        file is deleted after the operation.

        Parameters
        ----------
        output_file : str
            The path of the original output file.
        """

        # Log the refactoring process
        self.info("Refactoring output")

        # Split the output_file into filename and extension
        filename, file_extension = os.path.splitext(output_file)

        # Create a new name for the original file by appending '_unrefactored'
        new_name = "{}_unrefactored{}".format(filename, file_extension)

        # Rename the original file
        os.rename(output_file, new_name)

        # Open the renamed file and create a new output file
        # Then copy the contents from the renamed file to the new file
        with h5py.File(new_name, "r") as f:
            with h5py.File(output_file, "w") as g:
                copy_file(f, g)

        # Remove the renamed (old version) file
        os.remove(new_name)



import h5py
import numpy as np

from exosim.models import signal as signal
from exosim.tasks.load.loadOpticalElement import LoadOpticalElement
from exosim.utils import checks as checks


class LoadOpticalElementHDF5(LoadOpticalElement):
    """
    Class to load an optical element and return its radiance and efficiency using data from an HDF5 file.

    This class inherits from LoadOpticalElement and overrides the `model` and `_get_data` methods to use an HDF5 file
    for the optical element data.

    Returns
    --------
    :class:`~exosim.models.signal.Radiance`
        optical element radiance
    :class:`~exosim.models.signal.Dimensionless`
        optical element efficiency
    """

    def __init__(self):
        super().__init__()
        self.set_log_name()  # Ensure that the logger is properly initialized

    def _get_data(self, parameters, wl, tt, key, signal_type):
        """
        Load wavelength-dependent data from an HDF5 file, handling multiple structures.

        Parameters
        ----------
        parameters: dict
            dictionary containing the sources parameters.
        wl: :class:`~astropy.units.Quantity`
            wavelength grid.
        tt: :class:`~astropy.units.Quantity`
            time grid.
        key: str
            the key to read the specific dataset in the HDF5 group.
        signal_type: :class:`~exosim.models.signal.Signal`
            type of the signal to create.

        Returns
        -------
        :class:`~exosim.models.signal.Signal`
            extracted data as the specified signal type
        """
        hdf5_file_path = parameters["hdf5_file"]
        group_key = parameters["group_key"]
        read_key = parameters[key]

        with h5py.File(hdf5_file_path, "r") as hdf5_file:
            # Check if the group exists
            if group_key not in hdf5_file:
                raise KeyError(
                    f"Group '{group_key}' not found in HDF5 file '{hdf5_file_path}'"
                )

            group = hdf5_file[group_key]

            # Handle two possible structures
            if read_key in group and isinstance(group[read_key], h5py.Dataset):
                self.debug(
                    f"Reading data from dataset '{read_key}' in group '{group_key}'"
                )
                # Default structure: read_key and wavelength_key directly in group
                parsed_wl = checks.check_units(
                    group[parameters["wavelength_key"]][:], "um", force=True
                )
                parsed_data = group[read_key][:]
            elif read_key in group.keys() and isinstance(
                group[read_key], h5py.Group
            ):
                self.debug(
                    f"Reading data from folder '{read_key}' in group '{group_key}'"
                )
                # Alternative structure: read_key as a folder containing datasets
                subgroup = group[read_key]
                if (
                    parameters["wavelength_key"] not in subgroup
                    or read_key not in subgroup
                ):
                    raise KeyError(
                        f"Datasets '{parameters['wavelength_key']}' and '{read_key}' "
                        f"not found in folder '{read_key}' within group '{group_key}'"
                    )
                parsed_wl = checks.check_units(
                    subgroup[parameters["wavelength_key"]][:], "um", force=True
                )
                parsed_data = subgroup[read_key][:]
            else:
                raise KeyError(
                    f"Dataset or folder '{read_key}' not found in group '{group_key}'"
                )

        # Create and process the signal
        extracted_data = signal_type(spectral=parsed_wl, data=parsed_data)
        extracted_data.spectral_rebin(wl)
        extracted_data.temporal_rebin(tt)

        self.debug(
            f"Extracted data for key '{read_key}': {extracted_data.data}"
        )
        return extracted_data



import gc

import exosim.models.signal as signal
import exosim.tasks.instrument as instrument
from exosim.tasks.task import Task


class CreateFocalPlane(Task):
    """
    It produces the empty focal plane

    Returns
    -------
    :class:`~exosim.models.signal.Signal`
        focal plane array (with time evolution)
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        time: :class:`~astropy.units.Quantity`
            time grid.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        group_name: str (optional)
            group name in output
        """

        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("efficiency", "chennel efficiency")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file")
        self.add_task_param(
            "group_name", "group name in output", "focal plane"
        )

    def execute(self):
        self.info("creating focal plane")
        parameters = self.get_task_param("parameters")
        efficiency = self.get_task_param("efficiency")
        tt = self.get_task_param("time")
        output = self.get_task_param("output")
        group_name = self.get_task_param("group_name")

        createFocalPlaneArray = instrument.CreateFocalPlaneArray()
        focal_plane_array = createFocalPlaneArray(
            parameters=parameters, efficiency=efficiency
        )
        focal_plane_array.temporal_rebin(tt)

        focal_plane = signal.Signal(
            spectral=focal_plane_array.spectral,
            spatial=focal_plane_array.spatial,
            data=focal_plane_array.data,
            time=tt,
            cached=False,
            output=output,
            dataset_name=group_name,
            metadata=focal_plane_array.metadata,
        )

        self.set_output(focal_plane)

        del focal_plane
        gc.collect()



import numpy as np

from exosim.tasks.detector import LoadPixelsNonLinearityMap


class LoadPixelsNonLinearityMapNumpy(LoadPixelsNonLinearityMap):
    """
    Loads the pixels non-linearity map given a numpy map.
    The input must be a NPY format file (see `numpy documentation <https://numpy.org/devdocs/reference/generated/numpy.lib.format.html>`_) containing the coefficients for each pixel.
    The map should have dimensions (n, y, x) where n is the number of coefficients for the polynomial approximation and y and x are the number of pixels in spatial and spectral directions.
    The map should be indicated under ``pnl_filename`` keyword.

    Returns
    --------
    dict
        channel non linearity map

    Raises
    ------
    KeyError:
        if the output do not have the `map` key

    """

    def model(self, parameters: dict) -> dict:
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        dict
            channel pixel non-linearity map

        """
        file_name = parameters["detector"]["pnl_filename"]
        linear_coeff = np.load(file_name)

        pnl_dict = {"map": linear_coeff}

        return pnl_dict



from collections import OrderedDict

from exosim.tasks.astrosignal.estimateAstronomicalSignal import (
    EstimateAstronomicalSignal,
)
from exosim.tasks.task import Task
from exosim.utils.klass_factory import find_task


class FindAstronomicalSignals(Task):
    """
    This tasks find astronomical signals in the sky parameters dictionary.
    The signals are identified by the presence of the key "signal_task" in the dictionary.
    """

    def __init__(self) -> None:
        """
        Parameters
        ------------
        sky_parameters: dict
            sky parameters dictionary
        """
        self.add_task_param("sky_parameters", "sky parameters dictionary")

    def execute(self) -> None:
        self.sky_parameters = self.get_task_param("sky_parameters")
        self.signals_dict = {}
        if isinstance(self.sky_parameters["source"], OrderedDict):
            for source in self.sky_parameters["source"].keys():
                for effect in self.sky_parameters["source"][source].keys():
                    if isinstance(
                        self.sky_parameters["source"][source][effect], dict
                    ):
                        self.find_signals(
                            self.sky_parameters["source"][source][effect],
                            source,
                            effect,
                        )
        else:
            source = self.sky_parameters["source"]["value"]
            for effect in self.sky_parameters["source"].keys():
                if isinstance(self.sky_parameters["source"][effect], dict):
                    self.find_signals(
                        self.sky_parameters["source"][effect], source, effect
                    )
        if len(self.signals_dict) > 1:
            self.error(
                "Astronomical signals found in more than one star. "
                "The current version of ExoSim applies all the astronomical signals found in the target star. "
                "Please, check your sky parameters dictionary."
            )
            raise ValueError()

        self.set_output(self.signals_dict)

    def find_signals(self, input_dict: dict, source: str, effect: str) -> None:
        """
        Finds and stores the signal estimation task and parameters for a given source and effect.

        Parameters
        ----------
        input_dict : dict
            The dictionary containing the configuration parameters, including the optional key 'signal_task' which specifies the task responsible for signal estimation.
        source : str
            The name of the source for which the signals are to be found.
        effect : str
            The name of the effect to be considered in the signal estimation.

        """
        if "signal_task" in input_dict.keys():
            signal_task = find_task(
                input_dict["signal_task"], EstimateAstronomicalSignal
            )
            signal_dict = {
                "task": signal_task,
                "parsed_parameters": (
                    self.sky_parameters["source"][source]
                    if source in self.sky_parameters["source"].keys()
                    else self.sky_parameters["source"]
                ),
            }
            if source not in self.signals_dict.keys():
                self.signals_dict.update({source: {effect: signal_dict}})
            else:
                self.signals_dict[source].update({effect: signal_dict})



from tqdm.auto import tqdm

from exosim.models.signal import Adu
from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks


class MergeGroups(Task):
    """
    It averages the NDRs of the same group.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        n_groups: int
            number of groups in the same exposure
        n_ndrs: int
            number of ndrs to merge in the same group
        ndrs: :class:`~exosim.models.signal.Counts`
            ndrs cached signal'
        """
        self.add_task_param("subexposures", " ")
        self.add_task_param("n_groups", " ")
        self.add_task_param("n_ndrs", " ")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("merging groups")
        subexposures = self.get_task_param("subexposures")
        n_groups = self.get_task_param("n_groups")
        n_ndrs = self.get_task_param("n_ndrs")
        output = self.get_task_param("output")

        merged_ndrs = Adu(
            spectral=subexposures.spectral,
            time=subexposures.time[0::n_ndrs],
            data=None,
            spatial=subexposures.spatial,
            shape=(
                subexposures.dataset.shape[0] // n_ndrs,
                subexposures.dataset.shape[1],
                subexposures.dataset.shape[2],
            ),
            cached=True,
            output=output,
            dataset_name="NDRs",
            output_path=None,
            metadata=subexposures.metadata,
            dtype=subexposures.dataset[0, 0, 0].dtype,
        )

        if n_ndrs > 1:
            j = 0
            for i in tqdm(range(n_groups), desc="merging groups"):
                merged_ndrs.dataset[i] = (
                    subexposures.dataset[j : j + n_ndrs].sum(axis=0) / n_ndrs
                )
                j += n_ndrs
                merged_ndrs.output.flush()
        else:
            for chunk in iterate_over_chunks(
                subexposures.dataset, desc="merging groups"
            ):
                merged_ndrs.dataset[chunk] = subexposures.dataset[chunk]

        self.set_output(merged_ndrs)



from typing import TypeVar

import astropy.units as u
import numpy as np

ArrayType = TypeVar("ArrayType", np.ndarray, u.Quantity)
ValueType = TypeVar("ValueType", float, u.Quantity)
UnitType = TypeVar("UnitType", str, u.Quantity)
HDF5OutputType = TypeVar("HDF5OutputType", "HDF5Output", "HDF5OutputGroup")  # type: ignore # noqa: F821
OutputType = TypeVar("OutputType", "Output", "OutputGroup")  # type: ignore # noqa: F821



import numpy as np

from .exosimTool import ExoSimTool


class ADCGainEstimator(ExoSimTool):
    """
    It computes the desired ADC gain, given the number of bits used by the ADC
    and the maximum number of counts in the pixel to represent.

    Returns
    --------
    dict
        ADC gain factors

    Examples
    ----------

    >>> import exosim.tools as tools
    >>>
    >>> tools.ADCGainEstimator(options_file='tools_input_example.xml')
    """

    def __init__(self, options_file, output=None):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """
        super().__init__(options_file)

        self.info("creating dead pixel map")

        for ch in self.ch_list:
            gain_factor, dtype_max_val, int_type = self.model(
                self.ch_param[ch]
            )

            self.info("---------- {} --------".format(ch))
            self.info("gain factor: {}".format(gain_factor))
            self.info("max adc value: {}".format(dtype_max_val))
            self.info("integer dtype: {}".format(int_type))

            # prepare scheme output
            read_dict = {
                "gain factor": gain_factor,
                "max adc value": dtype_max_val,
                "integer dtype": int_type,
            }
            self.results.update({ch: read_dict})

    def model(self, parameters):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        float:
            adc gain factor
        int:
            max adc value
        str:
            data type used in ExoSim

        """
        self.info("converting to digital")

        if "ADC_num_bit" in parameters["detector"].keys():
            bits_num = parameters["detector"]["ADC_num_bit"]
            if isinstance(bits_num, float):
                if not bits_num.is_integer():
                    self.error("number of ADC bits should be integer")
                    raise TypeError("number of ADC bits should be integer")
            self.debug(
                "ADC set to {} bits.".format(
                    parameters["detector"]["ADC_num_bit"]
                )
            )
        else:
            bits_num = 32
            self.debug("ADC set to 32 bit by dfault.")

        if bits_num > 32:
            self.error("max bits number supported is 32.")
            raise ValueError("max bits number supported is 32.")

        dtype_max_val = 2**bits_num - 1

        int_type = np.dtype("int32")
        if bits_num <= 16:
            int_type = np.dtype("int16")
        if bits_num <= 8:
            int_type = np.dtype("int8")

        if "ADC_max_value" in parameters["detector"].keys():
            desired_max = parameters["detector"]["ADC_max_value"]
        else:
            desired_max = parameters["detector"]["well_depth"] * 1.1

        gain_factor = dtype_max_val / desired_max

        return gain_factor, dtype_max_val, int_type



from typing import Union

import astropy.units as u
import h5py
from astropy.io.misc.hdf5 import read_table_hdf5
from h5py import Dataset, Datatype, Group

import exosim.models.signal as signal


def load_signal(input: h5py.File) -> signal.Signal:
    """
    It loads the appropriate :class:`~exosim.models.signal.Signal` class from an opened hdf5 file.

    Parameters
    ----------
    input: :class:`h5py.File`
        opened hdf5 file

    Returns
    -------
    :class:`~exosim.models.signal.Signal`

    Raises
    -------
    IOError
        if the loaded Dataset is a Cached  :class:`~exosim.models.signal.Signal`.

    """
    metadata = {}
    if "metadata" in input.keys():
        metadata = recursively_read_dict_contents(input["metadata"])

    unit = u.Unit(input["data_units"][()].decode("utf-8"))

    spectral = (
        input["spectral"] * u.Unit(input["spectral_units"][()].decode("utf-8"))
        if "spectral" in input.keys()
        else None
    )

    time = (
        input["time"] * u.Unit(input["time_units"][()].decode("utf-8"))
        if "time" in input.keys()
        else None
    )

    spatial = (
        input["spatial"] * u.Unit(input["spatial_units"][()].decode("utf-8"))
        if "spatial" in input.keys()
        else None
    )

    # if 'datatype' in input.keys():
    #     class_name = input['datatype'][()].decode("utf-8").replace("<class '",
    #                                                                "").replace(
    #         "'>", "")
    #     klass = locate(class_name)
    # given the unit, check the right class to instantiate
    if unit == u.W / u.m**2 / u.um:
        klass = signal.Sed
    elif unit == u.W / u.m**2 / u.um / u.sr:
        klass = signal.Radiance
    elif unit == u.ct / u.s:
        klass = signal.CountsPerSecond
    elif unit == u.ct:
        klass = signal.Counts
    elif unit == u.adu:
        klass = signal.Adu
    elif unit == u.Unit(""):
        klass = signal.Dimensionless
    else:
        klass = signal.Signal

    if not input["cached"][()]:
        return klass(
            spectral=spectral,
            data=input["data"][()],
            data_units=unit,
            time=time,
            spatial=spatial,
            metadata=metadata,
        )
    else:
        raise OSError("impossible to load a cached dataset in a Signal class")


def recursively_read_dict_contents(input_dict: dict) -> dict:
    """
    Will recursive read a dictionary, initializing quantities and table from a dictionary read from an hdf5 file.

    Parameters
    ----------
    input_dict : dict
        dictionary read from hdf5

    Returns
    --------
    dict
        Dictionary we want to use

    """
    new_keys = [k for k in input_dict.keys()]
    output_dict = {}

    if all(elem in new_keys for elem in ["value", "unit"]):
        output_dict = input_dict["value"][()] * u.Unit(
            input_dict["unit"][()].decode("utf-8")
        )
        return output_dict

    elif any(".__table_column_meta__" in elem for elem in new_keys):
        table_keys = [
            elem for elem in new_keys if ".__table_column_meta__" in elem
        ]
        table_keys = (elem.split(".")[0] for elem in table_keys)
        for k in table_keys:
            table = read_table_hdf5(input_dict, k)
            output_dict[k] = table

    for key in new_keys:
        if isinstance(input_dict[key], h5py.Group):
            output_dict[key] = recursively_read_dict_contents(input_dict[key])
        elif isinstance(input_dict[key], h5py.Dataset):
            try:
                output_dict[key] = input_dict[key][()].decode("utf-8")
            except AttributeError:
                output_dict[key] = input_dict[key][()]

    return output_dict


def copy_file(
    in_object: Union[Group, Dataset], out_object: Union[Group, Dataset]
) -> None:
    """
    Recursively copy an HDF5 tree structure from one file to another.

    This function traverses the hierarchy of the input HDF5 object (`in_object`) and
    replicates it in the output HDF5 object (`out_object`). It can copy both groups and datasets,
    and also replicates all attributes.

    Parameters
    ----------
    in_object : Union[h5py.Group, h5py.Dataset]
        The input HDF5 object (either root, a subgroup, or a dataset).
    out_object : Union[h5py.Group, h5py.Dataset]
        The output HDF5 object (either root, a subgroup, or a dataset).

    Raises
    ------
    ValueError
        If an invalid object type is encountered.
    """

    # Copy attributes only once per group or dataset, to avoid overwriting
    for key, value in in_object.attrs.items():
        out_object.attrs[key] = value

    # Check if the input object is a dataset and skip further copying (base case for recursion)
    if isinstance(in_object, Dataset):
        return

    # Iterate through each key-value pair in the input object
    for key, in_obj in in_object.items():
        # Skip HDF5 Datatype objects
        if not isinstance(in_obj, Datatype):
            # Copy group
            if isinstance(in_obj, Group):
                out_obj = out_object.create_group(key)
                copy_file(in_obj, out_obj)  # Recursive call

            # Copy dataset
            elif isinstance(in_obj, Dataset):
                out_object.create_dataset(key, data=in_obj)

            # Raise an exception for unknown object types
            else:
                raise ValueError(f"Invalid object type {type(in_obj)}")



import os

import astropy.units as u
import h5py
import numpy as np

from exosim.tasks.task import Task
from exosim.utils.checks import check_units
from exosim.utils.types import ArrayType


class LoadILS(Task):
    """
    This Task loads the instrument line shapes (ILS) from the input file.
    The loaded ILS are then used to convolve the astronomical signal.
    The ILS here are intended as the PSF of the instrument.
    The ILS shapes are normalized to the maximum value.

    Returns
    -------
    :class:`numpy.ndarray`
        instrument line shape.
        The shape is (time, wavelength, spectral direction)

    Note
    -----
    The instrument line shapes produced by this task are not the same as the instrument line shapes
    as defined in the literature.
    The ILS produced by this task are the PSF of the instrument.
    To be used as the instrument line shapes as defined in the literature they need to be convolved with the intra-pixel response.
    This convolution is not part of this Task as it affects the way the ILS are sampled.
    The convolution with the intra-pixel response is done in the :class:`exosim.tasks.astrosignal.applyAstronomicalSignal.ApplyAstronomicalSignal` Task,
    where the ILS are used to convolve the astronomical signal.
    """

    def __init__(self):
        """
        Parameters
        -----------
        input_file: str
            focal plane input file
        ch_param: dict
            channel parameter
        wl_grid: :class:`astropy.units.Quantity`
            output wavelength grid
        """

        self.add_task_param("input_file", "focal plane input file")
        self.add_task_param("ch_param", "channel parameter")
        self.add_task_param("wl_grid", "wavelength grid")

    def execute(self):
        self.info("loading the instrument line shapes")
        parameters = self.get_task_param("ch_param")
        input_file = self.get_task_param("input_file")
        wl_grid = self.get_task_param("wl_grid")
        psf = self.model(input_file, parameters, wl_grid)
        self.set_output(psf)

    def model(
        self,
        input_file: str,
        parameters: dict,
        wl_grid: ArrayType,
    ) -> ArrayType:
        """
        It loads the channel instrument line shapes from the input file.

        Parameters
        ----------
        input_file: str
            focal plane input file
        ch_param: dict
            channel parameter
        wl_grid: :class:`astropy.units.Quantity`
            output wavelength grid

        Returns
        -------
        :class:`numpy.ndarray`
            instrument line shape.
            The shape is (time, wavelength, spectral direction)
        """
        ch_name = parameters["value"]

        with h5py.File(input_file, "r") as f:
            file_path = os.path.join("channels", ch_name)
            psf_path = os.path.join(file_path, "psf")
            psf_data = os.path.join(psf_path, "psf_cube")
            psf_wl = os.path.join(psf_path, "wavelength")

            psf = f[psf_data][()]
            sampled_wl = f[psf_wl]["value"][()] * u.Unit(f[psf_wl]["unit"][()])

        # select the wavelength range
        sampled_wl = check_units(sampled_wl, wl_grid.unit, self)
        sorter = np.argsort(sampled_wl)
        idx = np.searchsorted(sampled_wl.value, wl_grid.value, sorter=sorter)
        idx = idx[sorter]
        psf = psf[:, idx, :, :]
        # get the central slice
        central_spat = psf.shape[2] // 2
        # extract the central slice
        psf = psf[:, :, central_spat, :]
        # TODO what if the wl solution is curved? We need to interpolate
        # normalize the psf to the maximum
        psf /= psf.max(axis=-1)[:, :, np.newaxis]

        return psf



import logging

import numpy as np
from scipy.interpolate import interp1d
from scipy.stats import binned_statistic

from exosim.log import generate_logger_name


def rebin(
    x,
    xp,
    fp,
    axis=0,
    mode="mean",
    fill_value=0.0,
):
    """
    This function can resample multidimensional array along a given axis.
    Resample a function fp(xp) over the new grid x, rebinning if necessary,
    otherwise interpolates. Interpolation is done using 'linear' method.
    This function doesn't perform extrapolation: unsample coordinates will be filled with filled value.


    Parameters
    ----------
    x: 	:class:`~numpy.ndarray`
        New coordinates
    fp:	:class:`~numpy.ndarray`
        y-coordinates to be resampled
    xp:	:class:`~numpy.ndarray`
        x-coordinates at which fp are sampled
    axis: int (optional)
        fp axis to resample. Default is 0.
    mode: str (optional)
        the mode indicates the statistc to use for binning by :func:`scipy.stats.binned_statistic`.
        Default is 'mean'.
    fill_value: float or str (optional)
        fill value for unsampled coordinates. Default is 0.0.

    Returns
    -------
    :class:`~numpy.ndarray`
        re-sampled fp

    Raises
    -------
    NotImplementedError
        If the mode is not implemented.

    Examples
    ---------
    >>> import numpy as np
    >>> from exosim.utils.binning import rebin

    We define the original function, sampled in data 50 points:

    >>> xp = np.linspace(1,10, 50)
    >>> fp = np.sin(xp)

    We bin it down, sampling it at 10 points

    >>> x_bin = np.linspace(1,10, 10)
    >>> f_bin = rebin(x_bin, xp, fp)

    We wl_interpolate the original function, sampling it at 100 points

    >>> x_inter = np.linspace(1,10, 100)
    >>> f_inter = rebin(x_inter, xp, fp)


    To compare the outcome of our interpolation, we produce a plot.

    >>> import matplotlib.pyplot as plt
    >>> fig, (ax0, ax1) = plt.subplots(2,1)

    In the top panel we want to see the original function compared to the binned one and the interpolated one.

    >>> ax0.plot(xp, fp, label='original array', alpha=0.5, c='r')
    >>> ax0.scatter(x_inter, f_inter, marker='X', label='interpolated array', alpha=0.5)
    >>> ax0.scatter(x_bin, f_bin, marker='V', label='binned array', alpha=0.5)
    >>> ax0.legend()

    In the bottom panel we compare each function with the true value and we divide this quantity by the true value.

    >>> ax1.axhline(0, c='r', alpha=0.5)
    >>> ax1.scatter(x_inter, (f_inter- np.sin(x_inter))/np.sin(x_inter),
    ...             marker='X', label='interpolated array', alpha=0.5)
    >>> ax1.scatter(x_bin, (f_bin - np.sin(x_bin))/np.sin(x_bin),
    ...             marker='v', label='binned array', alpha=0.5)
    >>> ax1.legend()
    >>> plt.show()

    .. plot:: mpl_examples/rebin.py

    It's important to notice here that when we perform the binning, we are not simply resampling the input function,
    but we are using the mean value inside each of the bins.
    """

    x, xp, fp = _clean_up_inputs(x, xp, fp, axis)

    # binning
    idx = np.where(
        np.logical_and(x > np.min(xp), x < np.max(xp))
    )  # I'm only investigating in the interested area of the grid
    try:
        if (
            np.diff(xp).max() < np.diff(x[idx]).min()
        ):  # using max vs min I'm sure that if there is an empty bin it switches to interpolate
            return _bin_with_statistics(x, xp, fp, axis, statistic=mode)
        else:
            return _interpolating(x, xp, fp, axis, fill_value=fill_value)
    except ValueError:
        # if there are empty bins, switch to interpolation
        logger.warning("empty bins, switching to interpolation")
        return _interpolating(x, xp, fp, axis, fill_value=fill_value)


def _bin_with_statistics(x, xp, fp, axis, statistic="mean"):
    logger.debug("binning")

    bin_x = 0.5 * (x[1:] + x[:-1])
    x0 = x[0] - (bin_x[0] - x[0]) / 2.0
    x1 = x[-1] + (x[-1] - bin_x[-1]) / 2.0
    bin_x = np.insert(bin_x, [0], x0)
    bin_x = np.append(bin_x, x1)

    new_f = np.apply_along_axis(
        lambda a: binned_statistic(xp, a, bins=bin_x, statistic=statistic)[0],
        axis=axis,
        arr=fp,
    )
    return new_f


def _interpolating(x, xp, fp, axis, fill_value=0.0):
    logger.debug("interpolating")

    func = interp1d(
        xp,
        fp,
        axis=axis,
        fill_value=fill_value,
        assume_sorted=False,
        bounds_error=False,
        kind="linear",
    )
    new_f = func(x)
    return new_f


def _clean_up_inputs(x, xp, fp, axis):
    # cast into numpy array if they are not
    xp = np.array(xp)
    x = np.array(x)
    fp = np.array(fp)

    # remove NaNs
    idx = np.where(np.isnan(xp))[0]
    if idx.size > 0:
        logger.debug("Nans found in input x array: removing it")
        xp = np.delete(xp, idx)
        fp = np.delete(fp, idx, axis=axis)

    idx = np.where(np.isnan(fp))[0]
    if idx.size > 0:
        logger.debug("Nans found in input f array: removing it")
        xp = np.delete(xp, idx)
        fp = np.delete(fp, idx, axis=axis)

    idx = np.where(np.isnan(x))[0]
    if idx.size > 0:
        logger.debug("Nans found in new x array: removing it")
        x = np.delete(x, idx)

    # remove duplicates
    while np.diff(xp).min() == 0:
        logger.debug("duplicate found in input x array: removing it")
        idx = np.argmin(np.diff(xp))
        xp = np.delete(xp, idx)
        fp = np.delete(fp, idx, axis=axis)

    while np.diff(x).min() == 0:
        logger.debug("duplicate found in new x array: removing it")
        idx = np.argmin(np.diff(x))
        x = np.delete(x, idx)

    return x, xp, fp


logger = logging.getLogger(generate_logger_name(rebin))



import glob
import os

import numpy as np
from astropy import units as u
from astropy.io import fits

import exosim.models.signal as signal
import exosim.utils.checks as checks
from exosim.tasks.task import Task


class LoadPhoenix(Task):
    """
    Loads a star SED from a grid of Phoenix spectra or from a specific Phoenix file.

    Returns
    -------
    :class:`~exosim.models.signal.Sed`
        Star Sed

    Examples
    --------
    >>> from exosim.tasks.sed import LoadPhoenix
    >>> loadPhoenix = LoadPhoenix()

    Prepare the star

    >>> from astropy import constants as cc
    >>> import astropy.units as u
    >>> import numpy as np
    >>> D= 12.975 * u.pc
    >>> T= 3016 * u.K
    >>> M= 0.15 * u.Msun
    >>> R= 0.218 * u.Rsun
    >>> z= 0.0
    >>> g = (cc.G * M.si / R.si ** 2).to(u.cm / u.s ** 2)
    >>> logg = np.log10(g.value)

    Load the sed from a directory

    >>> sed = loadPhoenix(path = phoenix_directory,  T=T, D=D, R=R, z=z, logg=logg)

    or load the sed from a file

    >>> sed = loadPhoenix(filename = phoenix_file, D=D, R=R)

    Notes
    --------
    This class can either load the SED from a file or select the most suitable file from the Phoenix spectra path,
    given the proper information on the star. In the former case the `filename` keyword is needed, in the latter are
    required the keywords `path`, `T`, `z`, `logg`. Not providing the right keywords would result in an error.

    Raises
    --------
        InputError
            if neither `filename` or `path` are given.
        KeyError
            if a needed parameter is missing in the user inputs.

    """

    def __init__(self):
        """
        Parameters
        -----------
        R: :class:`~astropy.units.Quantity` or float
            star radius. If no units are attached is considered as expressed in `m`
        D: :class:`~astropy.units.Quantity` or float
            star distance. If no units are attached is considered as expressed in `m`
        filename: str (optional)
            Phoenix file path
        path: str (optional)
            Phoenix directory path
        T: :class:`~astropy.units.Quantity` or float (optional)
            star temperature. If no units are attached is considered as expressed in `K`
        z: float (optional)
            star metallicity.
        logg: float (optional)
            star logG.

        """
        self.add_task_param("R", "star radius")
        self.add_task_param("D", "star distannce")
        self.add_task_param("T", "star temperature", None)
        self.add_task_param("logg", "star logG", None)
        self.add_task_param("z", "star metallicity", None)
        self.add_task_param("path", "phoenix spectra path", None)
        self.add_task_param("filename", "phoenix file name", None)

    def execute(self):
        path = self.get_task_param("path")
        filename = self.get_task_param("filename")
        R = self.get_task_param("R")
        D = self.get_task_param("D")
        T = self.get_task_param("T")
        logg = self.get_task_param("logg")
        z = self.get_task_param("z")

        if not filename and not path:
            # TODO add Phoenix system path option if not path
            path = os.environ.get("PHOENIX_PATH", None)

        if (
            filename
        ):  # if filename is given it ignores anything else and it uses that
            fits_file_name = filename
        else:
            if path:
                if not os.path.exists(path):
                    # TODO test this
                    self.error(
                        "to load a phoenix model indicate a model file name or the phonix path"
                    )
                    raise OSError(
                        "to load a phoenix model indicate a model file name or the phonix path"
                    )
                if z is None:
                    z = 0.0
                    self.debug("star metallicity missing. zero is assumed.")
                if T is None:
                    self.error("star temperature missing")
                    raise KeyError("star temperature missing")
                if hasattr(T, "unit"):
                    T = T.to(u.K)
                else:
                    T = T * u.K
                    self.debug("no units found for T: Kelvin are assumed.")
                if not logg:
                    self.error("star logg missing")
                    raise KeyError("star logg missing")
                fits_file_name = self.get_phoenix_model_filename(
                    path, T, logg, z
                )
            else:
                self.error("phoenix path missing")
                raise OSError("phoenix path missing")
        sed = self.load(fits_file_name)
        sed.metadata["phoenix_file"] = fits_file_name
        if R is None:
            self.error("star radius missing")
            raise KeyError("star radius missing")
        if D is None:
            self.error("star distance missing")
            raise KeyError("star distance missing")

        R = checks.check_units(R, u.m, self)
        D = checks.check_units(D, u.m, self)

        sed *= (R / D) ** 2

        self.debug("phoenix sed loaded: {}".format(sed.data))

        self.set_output(sed)

    def load(self, ph_file):
        """
        Returs a sed given a Phoenix filename
        Parameters
        ----------
        ph_file: str
            filename

        Returns
        ---------
        :class:`~exosim.models.signal.Sed`
            Star Sed
        """
        with fits.open(ph_file) as hdu:
            str_unit = hdu[1].header["TUNIT1"]
            wl_ph = hdu[1].data.field("Wavelength") * u.Unit(str_unit)

            str_unit = hdu[1].header["TUNIT2"]
            sed_ph = hdu[1].data.field("Flux") * u.Unit(str_unit)

            idx = np.nonzero(np.diff(wl_ph))
            wl_ph = wl_ph[idx]
            sed_ph = sed_ph[idx]

        sed = signal.Sed(spectral=wl_ph, data=sed_ph)
        return sed

    def get_phoenix_model_filename(self, path, T, logg, z):
        """
        It returns the name of the phoenix file that best matches the input star parameters.

        Parameters
        ----------
        path: str
            Phoenix directory path
        T: :class:`~astropy.units.Quantity`
            star temperature
        logg: float
            star logG
        z: float
            star metallicity

        Returns
        -------
        str
            Phoenix file name
        """
        sed_name = glob.glob(os.path.join(path, "*.BT-Settl.spec.fits.gz"))

        if len(sed_name) == 0:
            self.error("No stellar SED files found")
            raise OSError("No stellar SED files found")

        sed_T_list = np.array(
            [float(os.path.basename(k)[3:8]) for k in sed_name]
        )
        sed_Logg_list = np.array(
            [float(os.path.basename(k)[9:12]) for k in sed_name]
        )
        sed_Z_list = np.array(
            [float(os.path.basename(k)[13:16]) for k in sed_name]
        )

        idx = np.argmin(
            np.abs(sed_T_list - np.round(T.value / 100.0))
            + np.abs(sed_Logg_list - logg)
            + np.abs(sed_Z_list - z)
        )

        ph_file = sed_name[idx]

        self.debug("phoenix file selected: {}".format(ph_file))
        return ph_file



import os

import astropy.units as u
import h5py
import numpy as np

import exosim.tasks.instrument as instrument
from exosim.output.hdf5.utils import load_signal
from exosim.utils import check_units

from .exosimTool import ExoSimTool


class ReadoutSchemeCalculator(ExoSimTool):
    """
    This tool helps in the definition of the channels' readout schemes.
    Starting from the desired readout format, this tool estimates
    the best parameters to set in the channel description to best fit the ramp.

    Attributes
    ----------
    input: str
        input file
    options: dict
        This is parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

    Examples
    --------
    >>> import exosim.tools as tools
    >>>
    >>> tools.ReadoutSchemeCalculator(options_file = 'tools_input_example.xml',
    >>>                               input_file='input_file.h5')
    """

    def __init__(self, options_file, input_file):
        super().__init__(options_file)
        self.input = input_file

        for ch in self.ch_list:
            focal_plane, frg_focal_plane = self.load_focal_plane(ch)
            self.info("Suggested readout scheme for {}".format(ch))
            (
                exposure_time,
                n_clk_GND,
                n_clk_ndr,
                n_NRDs_per_group,
                n_clk_GRP,
                n_GRPs,
                n_clk_RST,
            ) = self._compute_scheme(
                focal_plane, frg_focal_plane, self.ch_param[ch]
            )
            self._print_results(
                exposure_time,
                n_clk_GND,
                n_clk_ndr,
                n_NRDs_per_group,
                n_clk_GRP,
                n_GRPs,
                n_clk_RST,
            )
            # prepare scheme output
            read_dict = {
                "n_NRDs_per_group": n_NRDs_per_group,
                "n_groups": n_GRPs,
                "n_sim_clocks_Ground": n_clk_GND,
                "n_sim_clocks_Reset": n_clk_RST,
                "n_sim_clocks_first_NDR": n_clk_ndr,
                "n_sim_clocks_groups": n_clk_GRP,
                "exposure_time": exposure_time,
            }
            self.results.update({ch: read_dict})

    def _compute_scheme(self, focal_plane, frg_focal_plane, parameters):
        # compute saturation time
        computeSaturation = instrument.ComputeSaturation()
        (
            saturation_time,
            integration_time,
            max_signal,
            min_signal,
        ) = computeSaturation(
            well_depth=parameters["detector"]["well_depth"],
            f_well_depth=parameters["detector"]["f_well_depth"],
            focal_plane=focal_plane,
            frg_focal_plane=frg_focal_plane,
        )

        # load reading scheme parameters and convert to cycle clocks
        clock = check_units(parameters["readout"]["readout_frequency"], u.Hz)

        n_NRDs_per_group = parameters["readout"]["n_NRDs_per_group"]
        n_GRPs = parameters["readout"]["n_groups"]
        GND_time = parameters["readout"]["Ground_time"]
        RST_time = parameters["readout"]["Reset_time"]
        ndr_freq = check_units(
            parameters["readout"]["readout_frequency"], u.Hz
        )

        GND_time = check_units(GND_time, "s")
        n_clk_GND = int(np.ceil(GND_time * clock))
        RST_time = check_units(RST_time, "s")
        n_clk_RST = int(np.ceil(RST_time * clock))

        n_clk_ndr = int(np.ceil(clock / ndr_freq))

        if "exposure_time" in parameters["readout"].keys():
            exposure_time = parameters["readout"]["exposure_time"]
            exposure_time = check_units(exposure_time, "s")
        else:
            exposure_time = integration_time

        self.debug(
            "number of simulation clock in integration time: {}".format(
                exposure_time * clock
            )
        )
        available_clocks = int(
            np.floor(exposure_time * clock) - n_clk_RST - n_clk_GND - n_clk_ndr
        )

        # check for the number of groups
        clocks_per_group = (
            (n_NRDs_per_group - 1) * n_clk_ndr if n_NRDs_per_group > 1 else 1
        )
        n_clk_GRP = np.floor(
            available_clocks / (clocks_per_group * (n_GRPs - 1))
        ).astype(int)
        if n_clk_GRP < n_clk_ndr:
            while n_clk_GRP < n_clk_ndr:
                self.debug("reducing the number of groups")
                clocks_per_group = (
                    (n_NRDs_per_group - 1) * n_clk_ndr
                    if n_NRDs_per_group > 1
                    else 1
                )
                n_clk_GRP = np.floor(
                    available_clocks / (clocks_per_group * (n_GRPs - 1))
                ).astype(int)
                n_GRPs -= 1
                if n_GRPs <= 1:
                    self.warning(
                        "impossible to fit the desired number of groups inside the saturation time. Number of groups set to 1"
                    )
                    n_GRPs = 1
                    n_clk_GRP = 0
                    break

        # re-define exposures time
        exposure_time = self._get_exposure_time(
            n_clk_GND,
            n_clk_ndr,
            n_NRDs_per_group,
            n_clk_GRP,
            n_GRPs,
            n_clk_RST,
            clock,
        )
        if exposure_time > integration_time:
            self.warning(
                "Saturation problem: exposure time ({}) is {:.1f}% greater than saturation time ({}).".format(
                    exposure_time,
                    exposure_time / integration_time * 100,
                    integration_time,
                )
            )
            if "exposure_time" not in parameters["readout"].keys():
                while exposure_time > integration_time:
                    self.debug("reducing the number of NDRs")
                    exposure_time = self._get_exposure_time(
                        n_clk_GND,
                        n_clk_ndr,
                        n_NRDs_per_group,
                        n_clk_GRP,
                        n_GRPs,
                        n_clk_RST,
                        clock,
                    )
                    if n_NRDs_per_group == 1:
                        self.warning(
                            "impossible to fit the desired number of NDRs per group inside the saturation time. "
                            "Number of NDRs per group set to 1"
                        )
                        break
                    else:
                        n_NRDs_per_group -= 1
        return (
            exposure_time,
            n_clk_GND,
            n_clk_ndr,
            n_NRDs_per_group,
            n_clk_GRP,
            n_GRPs,
            n_clk_RST,
        )

    def _print_results(
        self,
        exposure_time,
        n_clk_GND,
        n_clk_ndr,
        n_NRDs_per_group,
        n_clk_GRP,
        n_GRPs,
        n_clk_RST,
    ):
        self.info("exposure time: {}".format(exposure_time))

        self.info("------------------------------------")
        self.info("n_NRDs_per_group:         {}".format(n_NRDs_per_group))
        self.info("n_GRPs:                   {}".format(n_GRPs))
        self.info("n_sim_clocks_Ground:      {}".format(n_clk_GND))
        self.info("n_sim_clocks_first_NDR:   {}".format(n_clk_ndr))
        self.info("n_sim_clocks_Reset:       {}".format(n_clk_RST))
        self.info("n_sim_clocks_groups:      {}".format(n_clk_GRP))
        self.info("------------------------------------")

    def _get_exposure_time(
        self,
        n_clk_GND,
        n_clk_ndr,
        n_NRDs_per_group,
        n_clk_GRP,
        n_GRPs,
        n_clk_RST,
        clock,
    ):
        return (
            n_clk_GND
            + n_clk_ndr * (n_NRDs_per_group)
            + (n_clk_GRP + n_clk_ndr * (n_NRDs_per_group - 1)) * (n_GRPs - 1)
            + n_clk_RST
        ) / clock.to(1 / u.s)

    def load_focal_plane(self, ch):
        """
        It loads the channel focal plane from the input file:

        Parameters
        ----------
        ch: str
            channel name

        Returns
        -------
        :class:`~exosim.models.signal.CountsPerSecond`
            focal plane
        :class:`~exosim.models.signal.CountsPerSecond`
            foreground focal plane
        """
        with h5py.File(self.input, "r") as f:
            file_path = os.path.join("channels", ch)
            fp = load_signal(f[os.path.join(file_path, "focal_plane")])
            frg_fp = load_signal(f[os.path.join(file_path, "frg_focal_plane")])

        self.debug("focal planes loaded from {}".format(ch))

        return fp, frg_fp



from exosim.tasks.task import Task


class ComputePhotonNoise(Task):
    r"""
    Computes the photon noise.
    Given the incoming signal :math:`S` the resulting photon noise variance is :math:`Var[S]=S`.
    If photon gain factor :math:`gain_{phot}` is given, then  :math:`Var[S]= gain_{phot} \cdot Var[S]`.
    If photon noise margin :math:`\chi` is found in the description, then  :math:`Var[S]= (1+\chi) \cdot Var[S]`.
    The noise returned is :math:`\sigma = \sqrt{Var[S]}`

    Returns
    -------
    astropy.table.QTable:
       photon noise
    """

    def __init__(self):
        """
        Parameters
        ----------
        signal: :class:`astropy.units.Quantity`
            signal
        description: dic (optional)
            channel description
        multiaccum_gain: :class:`numpy.ndarray` (optional)
            multiaccum gain factor for shotnoise
        """
        self.add_task_param("signal", "signal array")
        self.add_task_param("description", "channel description", None)
        self.add_task_param("multiaccum_gain", "channel description", None)

    def execute(self):
        self.debug("compute photon noise")
        signal = self.get_task_param("signal")
        description = self.get_task_param("description")
        multiaccum_gain = self.get_task_param("multiaccum_gain")

        variance = signal * signal.unit
        if description:
            if "photon_margin" in description["radiometric"].keys():
                self.debug(
                    "photon margin found: {}".format(
                        description["radiometric"]["photon_margin"]
                    ),
                )
                variance *= 1 + description["radiometric"]["photon_margin"]
        if multiaccum_gain is not None:
            variance *= multiaccum_gain

        self.set_output(variance**0.5)



import numpy as np
from astropy.io import ascii
from numba import jit, prange

from exosim.output import Output
from exosim.tasks.detector import ApplyDeadPixelsMap
from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks


class ApplyDeadPixelMapNumpy(ApplyDeadPixelsMap):
    """
    It masks the dead pixel in the array given a numpy map.
    The input must be a NPY format file (see `numpy documentation <https://numpy.org/devdocs/reference/generated/numpy.lib.format.html>`_) containing a boolean map marking with True the dead pixels.
    The map should be indicated under ``dp_map_filename`` keyword.
    """

    def model(self, subexposures, parameters, output):
        dead_pixels_map = np.ones(
            (subexposures.shape[1], subexposures.shape[2])
        )

        dead_map = np.load(parameters["detector"]["dp_map_filename"])

        if output:
            if issubclass(output.__class__, Output):
                out_grp = output.create_group("dead_pixel_map")
                out_grp.write_array("dead_pixel_map", dead_map)

        dead_pixels_map = np.invert(dead_map.astype(bool)).astype(int)

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="applying dead pixel map"
        ):
            subexposures.dataset[chunk] = self.add_dead_pixels(
                subexposures.dataset[chunk], dead_pixels_map
            )
            subexposures.output.flush()

    @staticmethod
    @jit(nopython=True, parallel=True)
    def add_dead_pixels(
        ndrs: np.ndarray, dead_pixels_map: np.ndarray
    ) -> np.ndarray:
        ndrs = ndrs.astype(np.float64)
        dead_pixels_map = dead_pixels_map.astype(np.float64)

        output = np.empty_like(ndrs)

        for t in prange(ndrs.shape[0]):
            output[t] = ndrs[t] * dead_pixels_map

        return output



import astropy.units as u
import h5py
import numpy as np
from astropy.table import QTable, Table

import exosim.models.signal as signal


def recursively_save_dict_contents_to_output(output, dic):
    """
    Will recursive write a dictionary into output.

    Parameters
    ----------
    :class:`~exosim.output.output.Output` :
        Group (or root) in output file to write to

    dic : :obj:`dict`
        Dictionary we want to write

    """

    for key, item in dic.items():
        try:
            store_thing(output, key, item)
        except TypeError:
            raise ValueError("Cannot write %s type" % type(item))
    return


def store_thing(output, key, item):
    """
    It stores one thing into the :class:`~exosim.output.output.Output`

    Parameters
    ----------
    :class:`~exosim.output.output.Output`:
        Group (or root) in output file to write to
    key: str
        name for the stored item
    item: obj
     item to store
    """
    if isinstance(item, u.Quantity):
        output.write_quantity(key, item)
    elif isinstance(
        item,
        (
            float,
            int,
            np.int64,
            np.float64,
        ),
    ):
        output.write_scalar(key, item)
    elif isinstance(item, np.ndarray):
        if True in [isinstance(x, str) for x in item]:
            output.write_string_array(key, item)
        else:
            output.write_array(key, item)
    elif isinstance(item, (str,)):
        output.write_string(key, item)
    elif isinstance(item, (Table, QTable)):
        output.write_table(key, item)
    elif isinstance(item, signal.Signal):
        item.write(output, key)
    #        group = output.create_group(key)
    #        recursively_save_dict_contents_to_output(group, item.to_dict())
    elif isinstance(
        item,
        (
            list,
            tuple,
        ),
    ):
        if isinstance(item, tuple):
            item = list(item)
        if True in [isinstance(x, str) for x in item]:
            output.write_string_array(key, np.array(item))
        else:
            try:
                output.write_array(key, np.array(item))

            except (TypeError, ValueError):
                for idx, val in enumerate(item):
                    new_key = "{}{}".format(key, idx)
                    store_thing(output, new_key, val)

    elif isinstance(item, dict):
        group = output.create_group(key)
        recursively_save_dict_contents_to_output(group, item)
    elif item is None:
        pass
    else:
        raise TypeError
    return



import os
import tempfile
from pathlib import Path

import h5py

import exosim.log as log

from .hdf5 import HDF5Output


class SetOutput(log.Logger):
    """
    It sets the output for the code.
    This class created and initializes the output file.
    If a file name is provided, it loads the relative :class:`~exosim.output.output.Output` class is instatiated.
    Otherwise an :class:`~exosim.output.hdf5.hdf5.HDF5Output` is used by default for a temporary file.
    """

    def __init__(self, filename: str = None, replace: bool = True):
        """

        Parameters
        ----------
        filename: str (optional)
            output file name for. If `None` a temporary file is produced.
        """
        super().__init__()

        self.tmp = False
        if filename is None:
            path = Path(os.path.join(os.getcwd(), "tmp"))
            tempfile.tempdir = path
            tempfile.tempdir.mkdir(parents=True, exist_ok=True)
            filename = tempfile.NamedTemporaryFile(
                suffix=".h5", delete=False
            ).name
            self.tmp = True
        self.info("file name set: {}".format(filename))
        dir_name = os.path.dirname(os.path.abspath(filename))
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
            self.debug("created {}".format(dir_name))

        self.fname = filename
        if replace:
            if os.path.exists(filename):
                os.remove(filename)

    def use(self, append: bool = True, cache: bool = False) -> HDF5Output:
        """
        It returns the :class:`~exosim.output.output.Output` with file opened and ready to write

        Parameters
        ----------
        append: bool (optional)
            True to append data to already existing file. Default is True.
        cache: bool (optional)
            True to write data in caching mode. Default is False.

        Returns
        -------
        :class:`~exosim.output.output.Output`
            output class instantiated.

        """
        if self.fname.endswith(".h5"):
            return HDF5Output(self.fname, append, cache)

    def open(self) -> h5py.File:
        """
        It returns the :class:`~exosim.output.output.Output` with file opened and ready to read

        Returns
        -------
        :class:`~exosim.output.output.Output`
            output class instantiated.
        """

        if self.fname.endswith(".h5"):
            f = HDF5Output(self.fname, append=True)
            f.open()
            return f.fd

    def __del__(self) -> None:
        """
        Garbage collector: it deletes the file when not in use by the context.
        """

        if self.tmp:
            self.delete()

    def delete(self) -> None:
        """
        It deletes the output file created.
        """

        try:
            os.remove(self.fname)
            self.debug("file deleted: {}".format(self.fname))
        except FileNotFoundError:
            pass

        # check if the file has been correctly removed
        if Path(self.fname).is_file():
            self.warning("file not deleted: {}".format(self.fname))

        # if temp dir is empty, delete it
        try:
            if not any(tempfile.tempdir.iterdir()):
                os.rmdir(tempfile.tempdir)
                self.debug(
                    "temporary dir deleted: {}".format(tempfile.tempdir)
                )
        except FileNotFoundError:
            pass
        except FileNotFoundError:
            pass
        except AttributeError:
            pass



from copy import deepcopy

import astropy.units as u
import numpy as np

from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks, searchsorted


class AddForegrounds(Task):
    """
    It adds the foregrounds the sub exposures.

    Returns
    --------
    :class:`~exosim.models.signal.Counts`
        sub-exposure cached signal class
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal class
        frg_focal_plane: :class:`~exosim.models.signal.Signal`
            channel foreground focal plane
        integration_time: :class:`~astropy.units.Quantity`
            sub-exposures integration times

        """

        self.add_task_param(
            "subexposures", "sub-exposures cached signal class"
        )
        self.add_task_param(
            "frg_focal_plane",
            "channel foreground focal plane",
        )
        self.add_task_param(
            "integration_time",
            "sub-exposures integration times",
        )

    def execute(self):
        self.info("adding foregrounds")
        subexposures = self.get_task_param("subexposures")
        frg_focal_plane = self.get_task_param("frg_focal_plane")
        integration_time = self.get_task_param("integration_time")

        osf = frg_focal_plane.metadata["oversampling"]

        ndrs_time = (subexposures.time * subexposures.time_units).to(u.hr)
        frg = deepcopy(frg_focal_plane.data.astype(np.float64))
        frg_time = (frg_focal_plane.time * frg_focal_plane.time_units).to(u.hr)
        index = searchsorted(frg_time, ndrs_time)
        self.debug("time indexes: {}".format(index))

        # bin down the frg to the subexposure size (integrate over pixels)
        frg = frg.reshape(
            (frg.shape[0], int(frg.shape[1] / osf), osf, frg.shape[2])
        ).sum(axis=2)
        frg = frg.reshape(
            (
                frg.shape[0],
                frg.shape[1],
                int(frg.shape[2] / osf),
                osf,
            )
        ).sum(axis=3)

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding foregrounds"
        ):
            subexposures.dataset[chunk] = self.add_frg(
                subexposures.dataset[chunk],
                frg,
                index[chunk[0]],
                integration_time[chunk[0]].to(u.s).value,
            )
            subexposures.output.flush()
        subexposures.metadata["frg_focal_plane_time_indexes"] = index
        self.set_output(subexposures)

    @staticmethod
    def add_frg(ndrs, frg, index, integration_time):
        for t in range(ndrs.shape[0]):
            ndrs[t] = ndrs[t] + frg[index[t]] * integration_time[t]
        return ndrs



from collections import OrderedDict

import astropy.units as u
import numpy as np
import requests
from astropy import constants as cc

import exosim.output as output
from exosim.tasks.sed import CreateCustomSource, PrepareSed
from exosim.tasks.task import Task
from exosim.utils.klass_factory import find_task


class ParseSources(Task):
    """
    Given the source descrition, it parses the sources elements and return a dictionary.
    It also applyes the time variation if provided.

    Returns
    -------
    dict
        dictionary containing :class:`~exosim.models.signal.Sed`

    Examples
    ---------

    >>> import astropy.units as u
    >>> import numpy as np
    >>> from exosim.tasks.parse import ParseSources
    >>> from collections import OrderedDict
    >>>
    >>> wl = np.linspace(0.5, 7.8, 10000) * u.um
    >>> tt = np.linspace(0.5, 1, 10) * u.hr
    >>>
    >>> sources_in = OrderedDict({'HD 209458': {'value': 'HD 209458',
    >>>                                         'source_type': 'planck',
    >>>                                         'R': 1.18 * u.R_sun,
    >>>                                         'D': 47 * u.pc,
    >>>                                         'T': 6086 * u.K,
    >>>                                         },
    >>>                           'GJ 1214': {'value': 'GJ 1214',
    >>>                                       'source_type': 'planck',
    >>>                                       'R': 0.218 * u.R_sun,
    >>>                                       'D': 13 * u.pc,
    >>>                                       'T': 3026 * u.K,
    >>>                                       }, })
    >>>
    >>> parseSources = ParseSources()
    >>> sources_out = parseSources(parameters=sources_in,
    >>>                            wavelength=wl,
    >>>                            time=tt)

    >>> import matplotlib.pyplot as plt
    >>>
    >>> plt.plot(source_out['HD 209458'].spectral, source_out['HD 209458'].data[0,0])
    >>> plt.ylabel(source_out['HD 209458'].data_units)
    >>> plt.xlabel(source_out['HD 209458'].spectral_units)
    >>> plt.show()

    .. plot:: mpl_examples/parseSources.py

    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        self.add_task_param("parameters", "sources parameters dict")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file", None)

    def execute(self):
        parameters = self.get_task_param("parameters")
        self.info("parsing sources")

        wl = self.get_task_param("wavelength")
        tt = self.get_task_param("time")

        output_file = self.get_task_param("output")

        parseSource = ParseSource()
        out = {}
        # if a list of sources are provided we parse all of them and we sum them up
        if isinstance(parameters, OrderedDict):
            for source_name in parameters.keys():
                source_ = parseSource(
                    parameters=parameters[source_name],
                    wavelength=wl,
                    time=tt,
                    output=output_file,
                )

                out = {**out, **source_}
        # else, we load the only source available
        else:
            out = parseSource(
                parameters=parameters,
                wavelength=wl,
                time=tt,
                output=output_file,
            )

        self.set_output(out)


class ParseSource(Task):
    """
    Given the source parameters, it parses the source element and returns a dictionary.
    It also applyes the time variation if provided.

    Returns
    -------
    dict
        dictionary containing :class:`~exosim.models.signal.Sed`

    Examples
    ---------

    >>> from exosim.tasks.parse import ParseSource
    >>> import astropy.units as u
    >>> import numpy as np
    >>> parseSource = ParseSource()
    >>> wl = np.linspace(0.5, 7.8, 10000) * u.um
    >>> tt = np.linspace(0.5, 1, 10) * u.hr
    >>> source_in = {
    >>>     'value': 'HD 209458',
    >>>     'source_type': 'planck',
    >>>     'R': 1.18 * u.R_sun,
    >>>     'D': 47 * u.pc,
    >>>     'T': 6086 * u.K,
    >>>     }
    >>> source_out = parseSource(parameters=source_in,
    >>>                          wavelength=wl,
    >>>                          time=tt)

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(source_out['HD 209458'].spectral, source_out['HD 209458'].data[0,0])
    >>> plt.ylabel(source_out['HD 209458'].data_units)
    >>> plt.xlabel(source_out['HD 209458'].spectral_units)
    >>> plt.show()

    .. plot:: mpl_examples/parseSource.py

    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        self.add_task_param("parameters", "sources parameters dict")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file", None)

    def execute(self):
        parameters = self.get_task_param("parameters")
        self.info("parsing source: {}".format(parameters["value"]))

        wl = self.get_task_param("wavelength")
        tt = self.get_task_param("time")

        sed = self._parse_sed(parameters, wl)

        sed.metadata["name"] = parameters["value"]
        sed.metadata["parsed_parameters"] = parameters

        sed.spectral_rebin(wl)
        sed.temporal_rebin(tt)

        output_file = self.get_task_param("output")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                og = output_file.create_group("sources")
                sed.write(og, sed.metadata["name"])

        out = {parameters["value"]: sed}
        self.set_output(out)

    def _parse_sed(self, parameters, wl):
        kwards = {"wavelength": wl}

        if "source_task" in parameters.keys():
            # if custom source is created by the user skip the automatic part
            source_task = find_task(
                parameters["source_task"], CreateCustomSource
            )
            task_instance = source_task()
            sed = task_instance(parameters=parameters)
            sed.metadata = {**sed.metadata, **kwards}
            return sed

        # initialize the keyword dictionary to the value into the xml or to None
        k_list = [
            "source_type",
            "R",
            "D",
            "logg",
            "T",
            "z",
            "path",
            "filename",
        ]
        for k in k_list:
            kwards[k] = None
            if k in list(parameters.keys()):
                kwards[k] = parameters[k]
                self.debug(
                    "source {} found in parameters: {}".format(
                        k, parameters[k]
                    )
                )

        # estimate the log from the mass if is not in the parameters but M and R are
        if "logg" not in list(parameters.keys()):
            try:
                g = (cc.G * parameters["M"].si / parameters["R"].si ** 2).to(
                    u.cm / u.s**2
                )
                kwards["logg"] = np.log10(g.value)
                self.debug("logg estimated: {}".format(kwards["logg"]))
            except KeyError:
                self.warning(
                    "Both mass (M) and radius (R) must be indicated to estimate logg"
                )

        # if an online data base is indicated it grab the data
        if "online_database" in list(parameters.keys()):
            # if parameters[
            #     'online_database'] == 'https://exodb.space/api/v1/star':
            response = requests.post(
                url=parameters["online_database"]["url"],
                headers=parameters["online_database"],
                json={"star_name": parameters["value"]},
            )
            star = response.json()["data"]["Properties"]
            kwards["R"] = star["Radius"]["value"] * u.Unit(
                star["Radius"]["unit"]
            )
            kwards["T"] = star["Effective Temperature"]["value"] * u.Unit(
                star["Effective Temperature"]["unit"]
            )
            kwards["D"] = star["Distance from Earth"]["value"] * u.Unit(
                star["Distance from Earth"]["unit"]
            )
            kwards["z"] = star["Metallicity"]["value"]
            M = star["Mass"]["value"] * u.Unit(star["Mass"]["unit"])
            g = (cc.G * M.si / kwards["R"].si ** 2).to(u.cm / u.s**2)
            kwards["logg"] = np.log10(g.value)
            self.debug("source data loaded from ExoDB")

        prepareSed = PrepareSed()
        sed = prepareSed(**kwards)
        sed.metadata = {**sed.metadata, **kwards}
        return sed



from importlib import import_module


def find_klass_in_file(python_file, baseclass):
    """
    It finds in the indicated python file a class that is a sublcass of the given one.

    Parameters
    ----------
    python_file: str
        python file name
    baseclass: class
        reference class to search for

    Returns
    -------
    class:
        class found in the python file.
    """

    import importlib.util
    import inspect

    spec = importlib.util.spec_from_file_location("foo", python_file)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)

    classes = [
        m[1]
        for m in inspect.getmembers(foo, inspect.isclass)
        if m[1] is not baseclass and issubclass(m[1], baseclass)
    ]

    if len(classes) == 0:
        # logger = logging.getLogger(generate_logger_name(obj))
        # logger.error('Could not find class of type %s in file %s',
        #           baseclass, python_file)
        raise Exception(
            f"No class inheriting from {baseclass} in " f"{python_file}"
        )
    return classes[0]


def load_klass(input, baseclass):
    """
    It returns a class that is a subclass of the given base class.

    Parameters
    ----------
    input: str or class
        if is a string, :func:`find_klass_in_file` is used to return the right class.
        If is a class, it checks whether it is an eligible class or not.
    baseclass: class
        reference class to search for

    Returns
    -------
    class:
        subclass of baseclass

    """

    if isinstance(input, str):
        return find_klass_in_file(input, baseclass)
    else:
        raise TypeError("task model in the wrong format")


def find_task(input, baseclass):
    """
    It looks for a class that is a subclass of the base class indicated.

    Parameters
    ----------
    input: str or object
        can either be a string indicating a class name, a python file, or it can be a class.

    baseclass: object
        reference class

    Returns
    -------
    object

    """
    if isinstance(input, str):
        if input == baseclass.__name__:
            # import the base class
            klass = baseclass
        elif input.endswith(".py"):
            # import from file
            klass = load_klass(input, baseclass)
        else:
            # look for task class by name in exosim.tasks submodules
            for sub in dir(import_module("exosim.tasks")):
                try:
                    klass = getattr(
                        import_module("exosim.tasks.{}".format(sub)), input
                    )
                except (ModuleNotFoundError, AttributeError):
                    continue
    elif issubclass(input, baseclass):
        klass = input
    else:
        raise TypeError
    return klass


def find_and_run_task(parameters, key, baseclass):
    """
    It looks in the input parameters for a class that is a subclass of the base class indicated, and it initialises it.

    Parameters
    ----------
    parameters: dict
        input dictionaty
    key: str
        string indicating the keyword for the class name

    baseclass: object
        reference class

    Returns
    -------
    callable
    """
    try:
        task = (
            find_task(parameters[key], baseclass)
            if key in parameters.keys()
            else baseclass
        )
    except UnboundLocalError as exc:
        raise Exception(
            "unable to find and instantiate a {} class".format(baseclass)
        ) from exc
    return task()



from .accumulateSubExposures import AccumulateSubExposures
from .addConstantDarkCurrent import AddConstantDarkCurrent
from .addCosmicRays import AddCosmicRays
from .addDarkCurrentMapNumpy import AddDarkCurrentMapNumpy
from .addGainDrift import AddGainDrift
from .addKTC import AddKTC
from .addReadNoise import AddNormalReadNoise
from .addReadNoiseMapNumpy import AddReadNoiseMapNumpy
from .addShotNoise import AddShotNoise
from .analogToDigital import AnalogToDigital
from .applyDeadPixelMap import ApplyDeadPixelsMap
from .applyDeadPixelMapNumpy import ApplyDeadPixelMapNumpy
from .applyPixelsNonLinearity import ApplyPixelsNonLinearity
from .applySimpleSaturation import ApplySimpleSaturation
from .loadPixelsNonLinearityMap import LoadPixelsNonLinearityMap
from .loadPixelsNonLinearityMapNumpy import LoadPixelsNonLinearityMapNumpy
from .mergeGroups import MergeGroups



import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d

import exosim.tasks.instrument as instrument
from exosim.tasks.task import Task


class SaturationChannel(Task):
    """
    It computes and adds the saturation time to the radiometric table using :class:`~exosim.tasks.instrument.computeSaturation.ComputeSaturation`.

    Returns
    -------
    astropy.units.Quantity:
        saturation time
    astropy.units.Quantity:
        maximum signal in focal plane
    astropy.units.Quantity:
        minimum signal in focal plane
    """

    def __init__(self):
        """
        Parameters
        --------------
        table: :class:`astropy.table.QTable`
            wavelength table with bin edge
        description: dict (optional)
            channel description
        input_file: :class:`~exosim.output.output.Output`
            input HDF5 file
        """
        self.add_task_param("table", "wavelength table with bin edges")
        self.add_task_param("description", "channel description", None)
        self.add_task_param(
            "input_file", "fraction of detector well depth to use"
        )

    def execute(self):
        input_file = self.get_task_param("input_file")
        description = self.get_task_param("description")
        table = self.get_task_param("table")

        ch = description["value"]
        computeSaturationTime = instrument.ComputeSaturation()
        with input_file.open() as f:
            f = f["channels"]
            focal_plane_units = u.Unit(f[ch]["focal_plane"]["data_units"][()])
            osf = f[ch]["focal_plane"]["metadata"]["oversampling"][()]
            focal_plane = (
                f[ch]["focal_plane"]["data"][
                    0, osf // 2 :: osf, osf // 2 :: osf
                ]
                * focal_plane_units
            )
            frg_focal_plane = (
                f[ch]["frg_focal_plane"]["data"][
                    0, osf // 2 :: osf, osf // 2 :: osf
                ]
                * focal_plane_units
            )

            sat_, int_, max_sig, min_sig = computeSaturationTime(
                well_depth=description["detector"]["well_depth"],
                f_well_depth=description["detector"]["f_well_depth"],
                focal_plane=focal_plane,
                frg_focal_plane=frg_focal_plane,
            )
            sat = [sat_] * len(table[table["ch_name"] == ch])
            integration_time = [int_] * len(table[table["ch_name"] == ch])

            # compute maximum signal in bin
            if description["type"].lower() == "spectrometer":
                wl_grid = f[ch]["focal_plane"]["spectral"][
                    osf // 2 :: osf
                ] * u.Unit(f[ch]["focal_plane"]["spectral_units"][()])
                wl_sol = interp1d(
                    wl_grid,
                    np.arange(0, focal_plane.shape[1]),
                    fill_value="extrapolate",
                )

                f = focal_plane + frg_focal_plane
                max_signal_in_bin, min_signal_in_bin = [], []
                for wl_l, wl_r in zip(
                    table["left_bin_edge"][table["ch_name"] == ch],
                    table["right_bin_edge"][table["ch_name"] == ch],
                ):
                    wld = int(wl_sol(wl_l))
                    wlu = int(wl_sol(wl_r))

                    if wld > wlu:
                        wld, wlu = wlu, wld
                    if wld < 0:
                        wld = 0
                    if wlu > f.shape[1]:
                        wlu = f.shape[1] - 1
                    max_signal_in_bin += [np.max(f[:, wld : wlu + 1])]
                    min_signal_in_bin += [np.min(f[:, wld : wlu + 1])]

            elif description["type"].lower() == "photometer":
                max_signal_in_bin = [np.max(focal_plane + frg_focal_plane)]
                min_signal_in_bin = [np.min(focal_plane + frg_focal_plane)]

        self.set_output(
            [sat, integration_time, max_signal_in_bin, min_signal_in_bin]
        )



import astropy.units as u
from joblib import Parallel, delayed

from exosim.tasks.task import Task
from exosim.utils import RunConfig


class PropagateSources(Task):
    """
    it propagates the sources though the channel.
    It multiplies the stellar SED by the effective telescope area, the channel efficiency and the channel responsivity:

    .. math::
        S_{\\nu} = S_{\\nu}^{*} \\times A_{tel} \\times \\eta \\times R_{\\nu}

    The result is in units of :math:`ct/s/\\mu m`

    Returns
    -------
    dict
        dictionary containing :class:`~exosim.models.signal.Signal`
    """

    def __init__(self):
        """
        Parameters
        __________
        sources:  dict
            dictionary containing :class:`~exosim.models.signal.Sed`
        Atel:  :class:`~astropy.units.Quantity`
            effective telescope Area
        efficiency:  :class:`~exosim.models.signal.Dimensionless`
            channel efficiency
        responsivity:  :class:`~exosim.models.signal.Signal`
            channel responsivity
        """

        self.add_task_param("sources", "sources dictionary")
        self.add_task_param("Atel", "effective telescope Area")
        self.add_task_param("efficiency", "channel efficiency")
        self.add_task_param("responsivity", "channel responsivity")

    def execute(self):
        sources = self.get_task_param("sources")
        Atel = self.get_task_param("Atel")
        efficiency = self.get_task_param("efficiency")
        responsivity = self.get_task_param("responsivity")

        Parallel(n_jobs=RunConfig.n_job, require="sharedmem")(
            delayed(self._propagate)(
                source, Atel, efficiency, responsivity, sources
            )
            for source in sources.keys()
        )
        # for source in sources.keys():
        #     sources[source] *= Atel * efficiency * responsivity
        #     sources[source].to(u.ct / u.s / u.um)

        self.set_output(sources)

    def _propagate(self, source, Atel, efficiency, responsivity, sources):
        sources[source] *= Atel * efficiency * responsivity
        sources[source].to(u.ct / u.s / u.um)



import astropy.units as u
import numpy as np

import exosim.output as output
from exosim.tasks.task import Task
from exosim.utils.checks import check_units


class ComputeReadingScheme(Task):
    """
    It computes the reading scheme for the sub-exposures given the instrument parameters and the focal planes.

    Returns
    -------
    :class:`astropy.units.Quantity`
        simulation frequency.
    :class:`numpy.ndarray`
        state machine for the reading operation on the ramp.
    :class:`numpy.ndarray`
        state machine of the group ends sampling the ramp.
    :class:`numpy.ndarray`
        state machine of the group starts sampling the ramp.
    :class:`numpy.ndarray`
        full list of simulation stapes for each steps on the ramp repeated by the number of ramps.
    int
        number of exposures needed to sample the full observation using ramps of the exposure time size.

    """

    def __init__(self):
        """
        Parameters
        ----------
        parameters: dict
            channel parameters dict
        main_parameters: dict
            main parameters dict
        output_file: :class:`~exosim.output.output.Output` (optional)
            output file

        """
        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("main_parameters", "main parameters dict")
        self.add_task_param(
            "readout_oversampling", "readout oversampling factor", 1
        )
        self.add_task_param("output_file", "output file")

    def execute(self):
        parameters = self.get_task_param("parameters")
        main_parameters = self.get_task_param("main_parameters")
        readout_oversampling = self.get_task_param("readout_oversampling")
        output_file = self.get_task_param("output_file")

        # load reading scheme parameters and convert to cycle clocks
        total_observing_time = main_parameters["time_grid"]["end_time"]
        self.info("total observing time: {}".format(total_observing_time))

        clock_freq = check_units(
            parameters["readout"]["readout_frequency"], u.Hz
        )
        clock_freq *= readout_oversampling
        clock = check_units(clock_freq, u.s)

        n_NRDs_per_group = parameters["readout"]["n_NRDs_per_group"]
        n_GRPs = parameters["readout"]["n_groups"]
        n_clk_GND = (
            parameters["readout"]["n_sim_clocks_Ground"] * readout_oversampling
        )
        n_clk_NDR0 = (
            parameters["readout"]["n_sim_clocks_first_NDR"]
            * readout_oversampling
        )
        n_clk_NDR = (
            parameters["readout"]["n_sim_clocks_NDR"]
            if "n_sim_clocks_NDR" in parameters["readout"].keys()
            else 1
        )
        n_clk_NDR *= readout_oversampling
        n_clk_RST = (
            parameters["readout"]["n_sim_clocks_Reset"] * readout_oversampling
        )
        n_clk_GRP = (
            parameters["readout"]["n_sim_clocks_groups"] * readout_oversampling
        )

        # define exposures time
        exposure_time = (
            n_clk_GND
            + n_clk_NDR0
            + n_clk_NDR * (n_NRDs_per_group - 1)
            + (n_clk_GRP + n_clk_NDR * (n_NRDs_per_group - 1)) * (n_GRPs - 1)
            + n_clk_RST
        ) * clock

        self.info("exposure time: {}".format(exposure_time))

        # define number of exposures
        if "n_exposures" in parameters["readout"].keys():
            number_of_exposures = parameters["readout"]["n_exposures"]
            self.debug(
                "number of exposures manually set to {}".format(
                    number_of_exposures
                )
            )
        else:
            number_of_exposures = np.ceil(
                total_observing_time.to(u.s) / exposure_time
            ).astype(int)
            self.debug(
                "number of exposures automatically set to {}".format(
                    number_of_exposures
                )
            )
        self.info("number of exposures: {}".format(number_of_exposures))

        # prepare scheme output
        read_dict = {
            "n_NRDs_per_group": n_NRDs_per_group,
            "n_GRPs": n_GRPs,
            "n_clk_GND": n_clk_GND,
            "n_clk_RST": n_clk_RST,
            "n_clk_ndr": n_clk_NDR,
            "n_clk_GRP": n_clk_GRP,
            "exposure_time": exposure_time,
            "number_of_exposures": number_of_exposures,
        }

        if issubclass(output_file.__class__, output.Output):
            output_file.store_dictionary(read_dict, "reading_scheme_params")

        # build state machines
        base = (
            [n_clk_GND]
            + [n_clk_NDR0]
            + [n_clk_NDR] * (n_NRDs_per_group - 1)
            + ([n_clk_GRP] + [n_clk_NDR] * (n_NRDs_per_group - 1))
            * (n_GRPs - 1)
            + [n_clk_RST]
        )
        base_mask = np.array([0] + [1] * (n_NRDs_per_group) * (n_GRPs) + [0])

        self.info(
            "number of NDRs: {}".format(number_of_exposures * base_mask.sum())
        )

        frame_sequence = np.tile(base, number_of_exposures)
        self.set_output(
            [clock, base_mask, frame_sequence, number_of_exposures]
        )



import numpy as np

import exosim.models.signal as signal
from exosim.tasks.subexposures import LoadQeMap


class LoadQeMapNumpy(LoadQeMap):
    """
    Loads the Quantum efficiency map from a NPY file (see `numpy documentation <https://numpy.org/devdocs/reference/generated/numpy.lib.format.html>`_).

    Returns
    --------
    :class:`~exosim.models.signal.Signal`
        channel responsivity variation map

    Raises
    ------
    TypeError:
        if the output is not a :class:`~exosim.models.signal.Signal` class

    """

    def model(self, parameters, time):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        time: :class:`~astropy.units.Quantity`
            time grid.

        Returns
        --------
        :class:`~exosim.models.signal.Signal`
            channel responsivity

        """
        file_name = parameters["detector"]["qe_map_filename"]
        qe_data = np.load(file_name)

        qe = signal.Signal(data=qe_data)

        return qe



from math import factorial
from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from tqdm.auto import tqdm

from exosim.output import SetOutput
from exosim.utils import RunConfig

from .exosimTool import ExoSimTool


class PixelsNonLinearity(ExoSimTool):
    r"""
    This tools helps the user to find the pixel non-linearity coefficients to as inputs for ExoSim,
    starting from the an estimate of pixel non-linearity correction.

    This class will retrieve the :math:`a_i` coefficients, starting from physical assumptions.

    The detector non linearity model, is written as polynomial such as

    .. math::
        Q_{det} = Q \cdot (1 + \sum_i a_i \cdot Q^i)

    where :math:`Q_{det}` is the charge read by the detector, and :math:`Q` is the ideal count,
    as :math:`Q = \phi_t`, with :math:`\phi` being the number of electrons generated and :math:`t` being the elapsed time.

    Considering the detector as a capacitor, the charge :math:`Q_{det}` is given by

    .. math::
        Q_{det} = \phi \tau \cdot \left(1 - e^{-Q/\phi \tau}\right)

    where :math:`\phi` is the charge generated in the detector pixel, and :math:`\tau` is the capacitor time constant.
    In fact the product :math:`\phi \tau` is constant
    :math:`Q` is the response of a linear detectror is given by :math:`Q = \phi t`

    The detector is considered saturated when the charge :math:`Q_{det}` at the well depth :math:`Q_{det, \, wd}`
    differs from the ideal well depth :math:`Q_{wd}` by 5%.

    .. math::
        Q_{det} = (1-5\%)Q_{wd}

    Then

    .. math::
        \phi \tau \cdot \left(1 - e^{-Q_{wd}/\phi \tau}\right) = (1-5\%)Q_{wd}

    This equation can be solved numerically and gives

    .. math::
        \frac{Q_{wd}}{\phi \tau} \sim 0.103479

    Therefore the detector collected charge is given by

    .. math::
            Q_{det} = \frac{Q_{wd}}{0.103479} \cdot \left(1 - e^\frac{- 0.103479 \, Q}{Q_{wd}}\right)

    Which can be approximated by a polynomial of order 4 as

    .. math::

            Q_{det} = Q\left[ 1- \frac{1}{2!}\frac{0.103479}{Q_{wd}} Q

            + \frac{1}{3!}\left(\frac{0.103479}{Q_{wd}}\right)^2 Q^2

            - \frac{1}{4!}\left(\frac{0.103479}{Q_{wd}}\right)^3 Q^3

            + \frac{1}{5!}\left(\frac{0.103479}{Q_{wd}}\right)^4 Q^4 \right]

    The results are the coefficients for a 4-th order polynomial:

    .. math::
        Q_{det} = Q \cdot (a_1 + a_2 \cdot Q + a_3 \cdot Q^2 + a_4 \cdot Q^3 + a_5 \cdot Q^4)

    However, each pixel is different, and therefore, this class also produces a map of the coefficient for each pixel.
    Each coefficient is normally distributed around the mean value, with a standard deviation indicated in the configuration.
    If no standard deviation is indicated, the coefficients are assumed to be constant.

    The code output is a map of :math:`a_i` coefficients for each pixel, which can be injected into  :class:`~exosim.tasks.detector.applyPixelsNonLinearity.ApplyPixelsNonLinearity`.


    Examples
    ----------

    >>> import exosim.tools as tools
    >>>
    >>> results = tools.PixelsNonLinearity(options_file='tools_input_example.xml',
    >>>                                    output='output_pnl_map.h5')
    """

    def __init__(
        self,
        options_file: Union[str, dict],
        output: str = None,
        show_results: bool = True,
    ) -> None:
        """

        Parameters
        ----------
        parameters: str or dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        output: str (optional)
            output file
        """
        super().__init__(options_file)
        if output is not None:
            out = SetOutput(output, replace=True)

        for ch in self.ch_list:
            out_dict = {}
            self.info(
                "computing pixel no linearity coefficients for {}".format(ch)
            )
            pnl_coeff, sat = self.compute_coefficients(
                self.ch_param[ch], show_results
            )
            out_dict["coeff"] = pnl_coeff
            out_dict["saturation"] = sat
            map = self.create_map(self.ch_param[ch], out_dict, show_results)
            out_dict["map"] = map
            self.results.update({ch: out_dict})

            if output is not None:
                with out.use(append=True) as o:
                    o.store_dictionary(out_dict, ch)

    def compute_coefficients(
        self, parameters: dict, show_results: bool = True
    ) -> Tuple[List[float], float]:
        """
        It computes the non linearity coefficients.

        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        show_results: bool
            it tells the code if showing the results in a plot. Default is `True`.
        """

        Q_wd = parameters["detector"]["well_depth"].value
        phi_t = 0.103479  # stimated solution of (1-exp(Qdet/phiT)) = (1-0.05)Qdet/phiT
        constant = phi_t / Q_wd
        # compute the coefficients
        pnl_coeff = np.array(
            [
                (-1) ** i * 1 / factorial(i + 1) * constant**i
                for i in range(1, 5)
            ]
        )

        if show_results:
            self._print_results(
                pnl_coeff,
                Q_wd,
            )
            self._plot(Q_wd, pnl_coeff)

        return np.insert(pnl_coeff, 0, 1), Q_wd

    def _plot(self, saturation, pnl_coeff) -> None:
        Q = np.linspace(
            1, saturation * 1.2, 2**10
        )  # detector pixel counts in adu

        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        fig.suptitle("pixel linearity model")
        ax.axvline(
            saturation,
            c="k",
            label="real saturation (5% from linear): {}".format(
                int(np.ceil(saturation))
            ),
            ls="-",
        )

        p = np.polynomial.Polynomial(np.insert(pnl_coeff, 0, 1))
        plt.plot(Q, Q * p(Q), "g", label="Detector pixel count")

        ax.plot(Q, Q, "k", ls=":", label="Linear pixel count")

        ax.set_xlabel("$Q$ [adu]")
        ax.set_ylabel("$Q_{det}$ [adu]")
        ax.legend()
        ax.grid()
        plt.show()

    def _print_results(self, coeff: List[float], sat: float) -> None:
        """It prints the estimated coefficients and saturation level to screen.

        Parameters
        ----------
        coeff : list
            list of fitted coefficents
        sat : float
            saturation value
        """

        self.info("saturation (5% from linear): {} (counts)".format(int(sat)))

        self.info("------------------------------------")
        self.info("pnl_coeff_a:     1")
        self.info("pnl_coeff_b:     {}".format(coeff[0]))
        self.info("pnl_coeff_c:     {}".format(coeff[1]))
        self.info("pnl_coeff_d:     {}".format(coeff[2]))
        self.info("pnl_coeff_e:     {}".format(coeff[3]))
        self.info("------------------------------------")

    def create_map(
        self, parameters: dict, input_dict: dict, show_results: bool = True
    ) -> np.ndarray:
        """
        Create a map of the pixel non-linearity correction coefficients.
        To create a non linearity map of the detector, we randomize the coefficients.
        If not specified, the standard deviation is set to 0 and the coefficients are assumed to be standard.
        of the mean value of the polynomial coefficients

        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters.
            This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        input_dict: dict
            dictionary produced by `compute_coefficients`.
            It contains the coefficients (`coeff`), the saturation ('saturation') and the estimated well depth (`well_depth`)

        show_results: bool
            it tells the code if showing the results in a plot. Default is `True`.


        Returns
        -------
        np.ndarray
            map of the coefficients. The shape is (4, nrow, ncol)
            The first axes refers to the coefficients (a, b, c, d, e)

        """

        map = np.ones(
            (
                input_dict["coeff"].size,
                parameters["detector"]["spatial_pix"],
                parameters["detector"]["spectral_pix"],
            )
        )

        if "pnl_coeff_std" in parameters["detector"]:
            std = parameters["detector"]["pnl_coeff_std"]
        else:
            std = 0.0

        for i in range(input_dict["coeff"].size):
            map[i] *= RunConfig.random_generator.normal(
                input_dict["coeff"][i],
                std * np.abs(input_dict["coeff"][i]),
                size=map.shape[1:],
            )

        if show_results:
            self._plot_map(input_dict, map)
        return map

    def _plot_map(self, input_dict, map):
        Q = np.linspace(
            1, input_dict["saturation"] * 1.2, 2**10
        )  # detector pixel counts in adu

        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        fig.suptitle("detector linearity model")
        ax.axvline(
            input_dict["saturation"],
            c="k",
            label="real saturation (5% from linear): {}".format(
                int(np.ceil(input_dict["saturation"]))
            ),
            ls="-",
        )

        coeffs = map.T.reshape(map.shape[1] * map.shape[2], map.shape[0])
        for cs in tqdm(coeffs, total=coeffs.shape[0], desc="preparing plot"):
            p = np.polynomial.Polynomial(cs)
            plt.plot(Q, Q * p(Q), c="g", lw=0.5, alpha=0.1)

        ax.plot(Q, Q, "k", ls=":", label="Linear pixel count")

        ax.set_xlabel("$Q$ [adu]")
        ax.set_ylabel("$Q_{det}$ [adu]")
        ax.legend()
        ax.grid()
        plt.show()



from typing import List

import numpy as np
from numba import jit

from exosim.tasks.task import Task
from exosim.utils.iterators import iterate_over_chunks


class AccumulateSubExposures(Task):
    """
    It accumulates sub-exposures of the same ramp.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        state_machine: :class:`numpy.ndarray`
            array indicating the exposures number of each sub-exposure.
            `
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("state_machine", "ramp state machine")

    def execute(self) -> None:
        self.info("accumulating sub-exposures")
        subexposures = self.get_task_param("subexposures")
        state_machine = self.get_task_param("state_machine")

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="accumulating sub-exposures"
        ):
            if (
                state_machine[chunk[0].start - 1]
                == state_machine[chunk[0].start]
            ):
                offset = subexposures.dataset[chunk[0].start - 1]
            else:
                offset = np.zeros_like(
                    subexposures.dataset[chunk[0].start - 1]
                )

            subexposures.dataset[chunk] = self.sub_exposures_cumsum(
                subexposures.dataset[chunk], state_machine[chunk[0]], offset
            )
            subexposures.output.flush()

    @staticmethod
    @jit(nopython=True)
    def sub_exposures_cumsum(
        dset: np.ndarray, state_machine: List[int], offset: List[int]
    ) -> np.ndarray:
        dset[0] += offset
        for i in range(1, dset.shape[0]):
            if state_machine[i] == state_machine[i - 1]:
                dset[i] += dset[i - 1]
        return dset



import astropy.units as u

from exosim.tasks.task import Task

from .aperturePhotometry import AperturePhotometry


class ComputeSignalsChannel(Task):
    """
    It estimates the radiometric signals on the input focal plane..

    Returns
    --------
    :class:`astropy.units.Quantity`
        photometry

    Raises
    --------
    TypeError:
        if the output is not :class:`~astropy.units.Quantity`
    UnitsError:
        wrong output units

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.

    """

    def __init__(self):
        """
        Parameters
        ----------
        table: :class:`astropy.table.QTable`
            apertures table
        focal plane: :class:`~exosim.output.output.Output`
            focal plane in the HDF5 file
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """
        self.add_task_param("table", "channel table")
        self.add_task_param("focal_plane", "focal_plane")
        self.add_task_param("parameters", "channel parameters dict", None)

    def execute(self):
        table = self.get_task_param("table")
        focal_plane = self.get_task_param("focal_plane")
        parameters = self.get_task_param("parameters")

        photometry = self.model(table, focal_plane, parameters)
        if not isinstance(photometry, u.Quantity):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        if not photometry.unit == u.ct / u.s:
            self.error("wrong output units")
            raise u.UnitsError("wrong output units")

        self.set_output(photometry)

    def model(self, table, focal_plane, parameters):
        """
        It estimates the radiometric signals on the input focal plane..
        It uses :func:`photutils.aperture.aperture_photometry` with the apertures
        from :class:`~exosim.tasks.radiometric.estimateApertures.EstimateApertures`.

        Parameters
        ----------
        table: :class:`astropy.table.QTable`
            apertures table
        focal plane: :class:`~exosim.output.output.Output`
            focal plane in the HDF5 file
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        :class:`astropy.units.Quantity`
            photometry

        """
        # prepare the focal plane
        self.debug("extracting focal plane. Oversampling removed")
        osf = focal_plane["metadata"]["oversampling"][()]
        focal_plane_units = u.Unit(focal_plane["data_units"][()])
        focal_plane = (
            focal_plane["data"][0, osf // 2 :: osf, osf // 2 :: osf]
            * focal_plane_units
        )

        aperturePhotometry = AperturePhotometry()

        self.debug("aperture photometry for source focal plane")
        signal_in_phot = aperturePhotometry(
            table=table, focal_plane=focal_plane
        )

        return signal_in_phot * focal_plane_units



from copy import deepcopy

import astropy.units as u
import numpy as np

from exosim.tasks.task import Task


class ComputeTotalNoise(Task):
    """
    It computes the total noise from a radiometric table

    Returns
    -------
    table: :class:`astropy.table.QTable`
        total noise column
    """

    def __init__(self):
        """
        Parameters
        --------------
        table: :class:`astropy.table.QTable`
            wavelength table with bin edge
        """
        self.add_task_param("table", "wavelength table with bin edges")

    def execute(self):
        table = self.get_task_param("table")

        total_variance = np.zeros(table["Wavelength"].size) * u.hr

        noise_k = [key for key in table.keys() if "noise" in key]
        for key in noise_k:
            # first iter on photon noise
            if "_photon_noise" in key and (
                "source" not in key and "foreground" not in key
            ):
                continue
            if table[key].unit == u.ct / u.s:
                contrib = deepcopy(
                    table[key].filled(0.0)
                    if hasattr(table[key], "filled")
                    else table[key]
                )
                contrib /= table["source_signal_in_aperture"]
                total_variance += (contrib**2 / u.hr.to(u.s)) * u.hr

        table["total_noise"] = (
            total_variance**0.5
        )  # / table['source_signal_in_aperture']

        total_variance = table["total_noise"] ** 2
        for key in noise_k:
            # second iter on relative noise
            if table[key].unit == u.hr**0.5 or table[key].unit is None:
                contrib = (
                    table[key].filled(0.0)
                    if hasattr(table[key], "filled")
                    else table[key]
                )
                total_variance += contrib**2 * u.hr
        table["total_noise"] = total_variance**0.5

        self.set_output(table["total_noise"])



import gc
import os
import re
from copy import deepcopy

import astropy.units as u
import h5py
import numpy as np

import exosim.tasks.instrument.loadPsfPaos as loadPsfPaos


class LoadPsfPaosTimeInterp(loadPsfPaos.LoadPsfPaos):
    """
    It loads the PSFs from two PAOS files and interpolates over time between them according to the given parameters.
    Each PAOS file contains a PSF vs wavelength.
    The task loads the PSF cubes provided by the `filename` parameter.
    The PSFs are then interpolated over a grid matching the one used to produce the focal planes,
    to convert them into the physical units.
    Then the total volume of the interpolated PSF is rescaled to the total volume of the original one.
    This allow to take into account for loss in the transmission due to the optical path.
    The PSF are then interpolated over a wavelength grid matching the one used to for the focal plane, producing the cube.

    Then an example of time dependency is implemented in the model method.

    The PSF cube is then cropped to removed unused edges, where the signal is less than 1/100 of the PSF peak.
    This would fasten up the successive `ExoSim` step.

    Returns
    -------
     :class:`~numpy.ndarray`
        cube of psfs.  axis=0 is time, axis=1 is wavelength, axis=2 is spatial direction, axis=3 is spectral direction.

    Note
    -----
    `filename` must be a string of filenames separated by a comma and a space.
    To write your version, define a new class that inherits from this one in a dedicated python file.
    Then copy the `model` method and replace the indicated section (where `REPLACE THIS WITH YOUR MODE` is written).

    """

    def model(self, filename, parameters, wavelength, time):
        self.debug("loading PAOS psf with time dependency")

        psfs = self.load_psfs(filename, wavelength, parameters)

        # prepare output ipercube
        psf_out = np.empty(
            (time.size, wavelength.size, psfs[0].shape[1], psfs[0].shape[2])
        )

        ################################
        # time evolution example.
        # REPLACE THIS WITH YOUR MODEL
        t0 = parameters["psf"]["t0"]
        for i, t in enumerate(time):
            psf_out[i] = (t / t0) * psfs[0] + (1 - t / t0) * psfs[1]

        ################################
        # clean memory
        del psfs
        gc.collect()

        # crop image stack
        psf_out_cropped = self.crop_image_stack(psf_out, time_iteration=True)

        # re-normalize output
        psf_out_cropped = self.normalise(psf_out, psf_out_cropped)

        return psf_out_cropped

    @staticmethod
    def normalise(psf_out, psf_out_cropped):
        old_norms = psf_out.sum(axis=-1).sum(axis=-1)
        new_norms = psf_out_cropped.sum(axis=-1).sum(axis=-1)
        dim_array = np.ones((1, psf_out.ndim), int).ravel()
        dim_array[0] = psf_out.shape[0]
        dim_array[1] = psf_out.shape[1]

        del psf_out  # clean memory
        gc.collect()

        old_norms_reshaped = old_norms.reshape(dim_array)
        new_norms_reshaped = new_norms.reshape(dim_array)
        psf_out_cropped_norm_ = psf_out_cropped / new_norms_reshaped

        del psf_out_cropped  # clean memory
        gc.collect()

        psf_out_cropped = psf_out_cropped_norm_ * old_norms_reshaped
        return psf_out_cropped

    def load_psfs(self, filename, wavelength, parameters):
        """

        Parameters
        ----------
         parameters: dict
             dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
         wavelength: :class:`~astropy.units.Quantity`
             wavelength grid.
         filename: str
             string containing PSF input files separated by a comma

        Returns
        -------
        list
            list of psf cubes. axis=0 is wavelength, axis=1 is spatial direction, axis=2 is spectral direction.


        """
        if "oversampling" in parameters["detector"].keys():
            oversampling = parameters["detector"]["oversampling"]
        else:
            oversampling = 1
        delta_pix = parameters["detector"]["delta_pix"] / oversampling
        filename = filename.split(", ")

        psfs = []
        for file in filename:
            with h5py.File(os.path.expanduser(file), "r") as data:
                # find the sampled wl
                temp = re.compile(r"\d+(?:\.\d*)")
                wl_sampled_h5 = [ele for ele in data.keys() if temp.match(ele)]

                psf_cube = [
                    self.load_imac(
                        wl, data, parameters, oversampling, delta_pix
                    )
                    for wl in wl_sampled_h5
                ]
                wl_sampled = [
                    self.load_wl_sampled(wl, data) for wl in wl_sampled_h5
                ]

            wl_sampled = np.array(wl_sampled) * u.m
            psf_cube = np.array(psf_cube)

            # wl_interpolate to the output wl grid
            wl_unsorted = deepcopy(wavelength)
            wavelength.sort()
            psf_cube_interp = self.wl_interpolate(
                psf_cube, wavelength, wl_sampled.to(u.um)
            )
            # reorder the array
            idx = [int(np.where(wavelength == wl_)[0]) for wl_ in wl_unsorted]
            psf_cube_interp = psf_cube_interp[idx, ...]

            psfs.append(psf_cube_interp)

        return psfs



import os
from copy import deepcopy
from typing import List, Tuple

import astropy.constants as const
import astropy.units as u
import h5py
import matplotlib
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec

import exosim.log as log
from exosim.output.hdf5.utils import load_signal
from exosim.utils.ascii_arts import observatory

from .utils import _create_ordered_cmap, prepare_channels_list

plt.rcParams.update({"font.size": 22})

cmap_band = _create_ordered_cmap("Pastel1", roll=-2, delete=-3)
cmap = _create_ordered_cmap("Set1", roll=-3, delete=2)


class FocalPlanePlotter(log.Logger):
    """
    Focal plane plotter.
    This class handles the methods to plot the focal palnes produced by `exosim`.

    Attributes
    -----------
    input: str
        input file name
    fig: :class:`matplotlib.figure.Figure`
        produced figure

    Examples
    ----------

    The following example, given the `test_file.h5` produced by `exosim`,
    plots the focal plane at the first time stamp and stores the figure as `focal_plane.png`.

    >>> from exosim.plots import FocalPlanePlotter
    >>> focalPlanePlotter = FocalPlanePlotter(input='./test_file.h5')
    >>> focalPlanePlotter.plot_focal_plane(time_step=0)
    >>> focalPlanePlotter.save_fig('focal_plane.png')

    """

    def __init__(self, input: str) -> None:
        """

        Parameters
        ----------
        input: str
            input data
        """
        self.set_log_name()
        self.graphics(observatory)

        self.input = input
        self.fig = None
        self.announce("started")

    def _prepare_figure(self) -> Tuple[plt.Figure, GridSpec, List[str]]:
        with h5py.File(self.input, "r") as f:
            ch_list = list(f["channels"].keys())
            ch_list.sort()
            widths = []
            for ch in ch_list:
                size_x = f["channels"][ch]["focal_plane"]["spectral"].shape[0]
                size_y = f["channels"][ch]["focal_plane"]["spatial"].shape[0]
                widths += [int(np.ceil(size_x / size_y)), 0.1]

            heights = [1, 1]
            scale = np.ceil(len(widths) / len(heights))
            size_y_fig = 10
            size_x_fig = size_y_fig * scale
            fig = plt.figure(
                constrained_layout=True,
                dpi=300,
                figsize=(size_x_fig, size_y_fig),
            )

            spec = fig.add_gridspec(
                ncols=len(widths),
                nrows=len(heights),
                width_ratios=widths,
                height_ratios=heights,
                wspace=0.1,
                hspace=0.1,
            )
        return fig, spec, ch_list

    def _plot_ch(
        self,
        ch: str,
        fig: plt.Figure,
        spec: GridSpec,
        i: int,
        time_step: int,
        scale="linear",
    ) -> Tuple[plt.Figure, int]:
        """
        It plots a channel focal plane in the indicated figure

        Parameters
        ----------
        ch: str
            channel name\
        fig: :class:`matplotlib.figure.Figure`
            figure to use
        spec: :class:`matplotlib.gridspec.GridSpec`
            figure grid spec
        i: int
            iterator
        time_step: int
            time step identifier
        scale: str (optional)
            scale to use for the plot. If 'dB' the plot is in dB. Default is 'linear'.

        Returns
        -------
        :class:`matplotlib.figure.Figure`
            populated figure
        i: int
           iterator
        """
        focal_plane, osf = self.load_focal_plane(ch, time_step)
        ax0 = fig.add_subplot(spec[0, i])
        ax1 = fig.add_subplot(spec[1, i])
        i += 1

        ax0.set_title(ch)

        ima = focal_plane
        ima = 10 * np.log10(ima / ima.max()) if scale == "dB" else ima
        im = ax0.imshow(ima, interpolation="none")
        plt.colorbar(im, ax=ax0, cax=fig.add_subplot(spec[0, i]))

        ax1.set_title("extracted focal plane")

        ima = focal_plane[osf // 2 :: osf, osf // 2 :: osf]
        ima = 10 * np.log10(ima / ima.max()) if scale == "dB" else ima
        im = ax1.imshow(ima, interpolation="none")
        plt.colorbar(im, ax=ax1, cax=fig.add_subplot(spec[1, i]))
        return fig, i

    def load_focal_plane(
        self, ch: str, time_step: int
    ) -> Tuple[np.ndarray, int]:
        """
        It loads the channel focal plane from the input file:

        Parameters
        ----------
        ch: str
            channel name
        time_step: int
            time step to plot

        Returns
        -------
        :class:`numpy.ndarray`
            focal plane
        int
            over sampling factor
        """
        with h5py.File(self.input, "r") as f:
            file_path = os.path.join("channels", ch)
            source = load_signal(f[os.path.join(file_path, "focal_plane")])
            frg = load_signal(
                f[os.path.join(file_path, "frg_focal_plane")]
            ).data[time_step]
            try:
                bkg = load_signal(
                    f[os.path.join(file_path, "bkg_focal_plane")]
                ).data[time_step]
            except KeyError:
                bkg = 0
            osf = source.metadata["oversampling"]

            final = source.data[time_step] + frg + bkg

        self.debug("focal planes loaded from {}".format(ch))

        return final, osf

    def plot_focal_plane(
        self, time_step: int = 0, scale="linear"
    ) -> plt.Figure:
        """
        It plots the focal planes at a specific time.
        For each channel it adds a :class:`~matplotlib.axes.Axes` to a figure.
        It returns a :class:`~matplotlib.figure.Figure` with two rows:
        on the first row are reported the oversampled focal planes.
        In the second row are reported the extracted focal plane,
        where the oversampling is removed.
        The focal plane plotted is the combination of the source focal plane
        plus the foreground focal plane.

        Parameters
        ----------
        time_step: int
            time step identifier
        scale: str (optional)
            scale to use for the plot. If 'dB' the plot is in dB. Default is 'linear'.

        Returns
        -------
        :class:`matplotlib.figure.Figure`
            populated figure
        """
        self.info("plot focal plane")

        fig, spec, ch_list = self._prepare_figure()
        i = 0
        for ch in ch_list:
            fig, i = self._plot_ch(ch, fig, spec, i, time_step, scale)
            i += 1

        self.fig = fig

        return fig

    def plot_bands(
        self,
        ax: matplotlib.axes.Axes,
        scale: str = "log",
        channel_edges: bool = True,
        add_legend: bool = True,
    ) -> matplotlib.axes.Axes:
        """
        It plots the channels bands behind the indicated axes.

        Parameters
        -----------
        ax: :class:`matplotlib.axes.Axes`
            axes where to plot the bands
        scale: str
            x axes scale. Default is `log`.
        channel_edges: bool
            if ``True`` the x axes ticks are placed at the channel edges. Default is ``True``.

        Returns
        --------
        :class:`matplotlib.axes.Axes`
            axes with channel bands added

        """
        channels, norm = prepare_channels_list(self.input)

        tick_list, patches = [], []
        for k, channel_name in enumerate(channels):
            with h5py.File(self.input, "r") as f:
                file_path = os.path.join("channels", channel_name)
                file_path = os.path.join(file_path, "focal_plane/metadata")
                wl_min = f[os.path.join(file_path, "wl_min/value")][()]
                wl_max = f[os.path.join(file_path, "wl_max/value")][()]

                ax.axvspan(
                    wl_min,
                    wl_max,
                    alpha=0.3,
                    zorder=0,
                    color=cmap_band(
                        norm(k),
                    ),
                )
                ax.axvspan(
                    wl_min,
                    wl_max,
                    alpha=0.3,
                    zorder=0,
                    color=cmap_band(
                        norm(k),
                    ),
                )
                #            wl_maxs += [wl_max]
                tick_list.append(wl_min)
                tick_list.append(wl_max)
                patches += [
                    mpatches.Patch(
                        color=cmap_band(norm(k)), alpha=0.1, label=channel_name
                    )
                ]
            #       tick_list.append(max(wl_maxs))
        ax.set_xscale(scale)
        if channel_edges:
            ax.set_xticks(tick_list)
            ax.get_xaxis().set_major_formatter(
                matplotlib.ticker.ScalarFormatter()
            )

        if add_legend:
            band_legend = ax.legend(
                handles=patches, title="Channels legend", bbox_to_anchor=(1, 1)
            )
            ax.add_artist(band_legend)
        return ax

    def plot_efficiency(
        self,
        scale: str = "log",
        channel_edges: bool = False,
        ch_lengend: bool = False,
        efficiency: str = "all",
    ) -> Tuple[plt.Figure, List[matplotlib.axes.Axes]]:
        """
        It produces a figure with efficiencies for the input table.

        Parameters
        ----------
        scale: str
            x axes scale. Default is `log`.
        channel_edges: bool
            if ``True`` the x axes ticks are placed at the channel edges. Default is ``True``.
        ch_lengend: bool
            if ``True`` add a legend for the channels color. Default is ``True``.
        efficiency: str
            what efficiency to plot. Options are "optical efficiency", "responsivity", "quantum efficiency", "photon conversion efficiency" and "all". Default is "all".
        Returns
        --------
        :class:`matplotlib.figure.Figure`
            plotted figure
        (:class:`matplotlib.axes.Axes`, :class:`matplotlib.axes.Axes`, :class:`matplotlib.axes.Axes`, :class:`matplotlib.axes.Axes`)
            tuple of axis. First axes is for optical efficiency, second is for responsivity, third is for quantum efficiency and fourth is for photon conversion efficiency.

        """
        self.info("plotting efficiency")

        channels, norm = prepare_channels_list(self.input)

        def bands(axs):
            try:
                for ax in axs.flatten():
                    self.plot_bands(
                        ax, scale, channel_edges, add_legend=ch_lengend
                    )
            except:
                try:
                    axs = self.plot_bands(
                        axs, scale, channel_edges, add_legend=ch_lengend
                    )
                except KeyError:
                    pass

        with h5py.File(self.input, "r") as f:
            if efficiency == "optical efficiency":
                fig, axs = plt.subplots(1, 1, figsize=(10, 8))
                bands(axs)
                for k, ch in enumerate(channels):
                    eff_path = "channels/{}/efficiency".format(ch)
                    eff = load_signal(f[eff_path])
                    axs.plot(
                        eff.spectral,
                        eff.data[0, 0],
                        label=ch,
                        color=cmap(norm(k)),
                        zorder=0,
                    )
                axs.set_title("Efficiency")
                axs.set_ylabel("Efficiency")
                axs.set_xlabel(r"Wavelength [$\mu m$]")
                axs.grid()

            elif efficiency == "responsivity":
                fig, axs = plt.subplots(1, 1, figsize=(10, 8))
                bands(axs)
                for k, ch in enumerate(channels):
                    resp_path = "channels/{}/responsivity".format(ch)
                    resp = load_signal(f[resp_path])
                    axs.plot(
                        resp.spectral,
                        resp.data[0, 0],
                        label=ch,
                        color=cmap(norm(k)),
                        zorder=0,
                    )
                axs.set_title("Responsivity")
                axs.set_ylabel(r"${}$".format(resp.data_units))
                axs.set_xlabel(r"Wavelength [$\mu m$]")
                axs.grid()

            elif efficiency == "quantum efficiency":
                fig, axs = plt.subplots(1, 1, figsize=(10, 8))
                bands(axs)
                for k, ch in enumerate(channels):
                    resp_path = "channels/{}/responsivity".format(ch)
                    resp = load_signal(f[resp_path])
                    wl = resp.spectral * u.Unit(resp.spectral_units)
                    qe = (
                        resp.data[0, 0]
                        / wl.to(u.m)
                        * const.c
                        * const.h
                        / u.count
                        * resp.data_units
                    )
                    axs.plot(
                        resp.spectral,
                        qe,
                        label=ch,
                        color=cmap(norm(k)),
                        zorder=0,
                    )
                axs.set_title("Quantum efficiency")
                axs.set_ylabel("Efficiency")
                axs.set_xlabel(r"Wavelength [$\mu m$]")
                axs.grid()

            elif efficiency == "photon conversion efficiency":
                fig, axs = plt.subplots(1, 1, figsize=(10, 8))
                bands(axs)
                for k, ch in enumerate(channels):
                    eff_path = "channels/{}/efficiency".format(ch)
                    eff = load_signal(f[eff_path])
                    resp_path = "channels/{}/responsivity".format(ch)
                    resp = load_signal(f[resp_path])
                    wl = resp.spectral * u.Unit(resp.spectral_units)
                    qe = (
                        resp.data[0, 0]
                        / wl.to(u.m)
                        * const.c
                        * const.h
                        / u.count
                        * resp.data_units
                    )
                    axs.plot(
                        resp.spectral,
                        qe * eff.data[0, 0],
                        label=ch,
                        color=cmap(norm(k)),
                        zorder=0,
                    )
                axs.set_title("photon conversion efficiency")
                axs.set_ylabel("Efficiency")
                axs.set_xlabel(r"Wavelength [$\mu m$]")
                axs.grid()

            elif efficiency == "all":
                fig, axs = plt.subplots(2, 2, figsize=(20, 15))
                bands(axs)

                ax1, ax2, ax3, ax4 = axs.flatten()
                with h5py.File(self.input, "r") as f:
                    for k, ch in enumerate(channels):
                        eff_path = "channels/{}/efficiency".format(ch)
                        eff = load_signal(f[eff_path])
                        ax1.plot(
                            eff.spectral,
                            eff.data[0, 0],
                            label=ch,
                            color=cmap(norm(k)),
                            zorder=0,
                        )

                        resp_path = "channels/{}/responsivity".format(ch)
                        resp = load_signal(f[resp_path])
                        ax2.plot(
                            resp.spectral,
                            resp.data[0, 0],
                            label=ch,
                            color=cmap(norm(k)),
                            zorder=0,
                        )

                        wl = resp.spectral * u.Unit(resp.spectral_units)
                        qe = (
                            resp.data[0, 0]
                            / wl.to(u.m)
                            * const.c
                            * const.h
                            / u.count
                            * resp.data_units
                        )
                        ax3.plot(
                            resp.spectral,
                            qe,
                            label=ch,
                            color=cmap(norm(k)),
                            zorder=0,
                        )

                        ax4.plot(
                            resp.spectral,
                            qe * eff.data[0, 0],
                            label=ch,
                            color=cmap(norm(k)),
                            zorder=0,
                        )

                # locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
                # ax.yaxis.set_major_locator(locmaj)
                # locmin = matplotlib.ticker.LogLocator(base=10.0,
                #                                       subs=(0.2, 0.4, 0.6, 0.8),
                #                                       numticks=12)
                # ax.yaxis.set_minor_locator(locmin)
                # ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
                # ax.grid(axis='y', which='minor', alpha=0.3)
                # ax.grid(axis='y', which='major', alpha=0.5)
                # #        ax.legend(bbox_to_anchor=(1, 1))
                ax1.set_title("Efficiency")
                ax1.set_xlabel(r"Wavelength [$\mu m$]")
                ax1.set_ylabel("Efficiency")
                ax1.grid()

                ax2.set_title("Responsivity")
                ax2.set_ylabel(r"${}$".format(resp.data_units))
                ax2.set_xlabel(r"Wavelength [$\mu m$]")
                ax2.set_yscale("log")
                ax2.grid()

                ax3.set_title("Quantum efficiency")
                ax3.set_xlabel(r"Wavelength [$\mu m$]")
                ax3.set_ylabel("Efficiency")
                ax3.grid()

                ax4.set_title("Photon conversion efficiency")
                ax4.set_xlabel(r"Wavelength [$\mu m$]")
                ax4.set_ylabel("Efficiency")
                ax4.grid()

            try:
                ax3.legend(
                    prop={"size": 16},
                    loc="upper left",
                    ncol=7,
                    bbox_to_anchor=(0.05, -0.11),
                    labelspacing=1.2,
                    handlelength=1,
                )
                plt.tight_layout()
                plt.subplots_adjust(
                    top=0.92,
                    bottom=0.12,
                    hspace=0.2,
                    wspace=0.2,
                    right=0.85,
                    left=0.1,
                )
            except:
                axs.legend(
                    prop={"size": 12},
                    loc="upper left",
                    ncol=7,
                    bbox_to_anchor=(0.05, -0.11),
                    labelspacing=1.2,
                    handlelength=1,
                )
                plt.tight_layout()
                plt.subplots_adjust(
                    top=0.92,
                    bottom=0.15,
                    # hspace=0.2,
                    # wspace=0.2,
                    # right=0.9,
                    # left=0.1,
                )

        self.fig = fig
        return fig, axs

    def save_fig(self, name: str) -> None:
        """
        It saves the produced figure.

        Parameters
        --------
        name: str
            figure name
        """
        dir_name = os.path.dirname(os.path.abspath(name))
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)

        try:
            self.fig.savefig("{}".format(name))

            self.info("plot saved in {}".format(name))
        except AttributeError:
            self.error(
                "the indicated figure is not available. Check if you have produced it."
            )



import logging

import numpy as np
from scipy.special import j1, jn_zeros

logger = logging.getLogger(__name__)


def create_psf(
    wl,
    fnum,
    delta,
    nzero=4,
    shape="airy",
    max_array_size=None,
    array_size=None,
):
    """
    Calculates an Airy Point Spread Function arranged as a data-cube.
    The spatial axes are 0 and 1. The wavelength axis is 2.
    Each PSF volume is normalised to unity.

    Parameters
    ----------
    wl: :class:`~astropy.units.Quantity`
        array of wavelengths at which to calculate the PSF
    fnum: float or (float, float)
        Instrument f/number. It can be a tuple of two values, in which case the first value is the f/number for the x axis and the second value is the f/number for the y axis.
    delta: :class:`~astropy.units.Quantity`
        the increment to use [physical unit of length]
    nzero: float
        number of Airy zeros. The PSF kernel will be this big. Calculated at wl.max()
    shape: str (optional)
        Set to 'airy' for a Airy function,to 'gauss' for a Gaussian
    max_array_size: (int,int) (optional)
        Maximum size of the PSF array. If None, the size is calculated from the f/number and the wavelength range.
    array_size: (int or str,int or str) (optional)
        Size of the PSF array. If 'full' then the `max_array_size` are used. If None, the size is calculated from the f/number and the wavelength range.

    Returns
    ------
    :class:`~numpy.ndarray`
        three-dimensional array. Each PSF normalised to unity

    Examples
    ---------
    >>> import astropy.units as u
    >>> from exosim.utils.psf import create_psf

    We produce and plot an Airy PSF:

    >>> img = create_psf(1*u.um, 40, 6*u.um, nzero=8, shape='airy')

    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from matplotlib.gridspec import GridSpec
    >>> fig = plt.figure(figsize=(6, 6))
    >>> gs = fig.add_gridspec(2, 2,  width_ratios=(4, 1), height_ratios=(1, 4),
    >>>                       left=0.1, right=0.9, bottom=0.1, top=0.9,
    >>>                       wspace=0.05, hspace=0.05)
    >>> ax = fig.add_subplot(gs[1, 0])
    >>> ax.imshow(img)
    >>> ax_x = fig.add_subplot(gs[0, 0], sharex=ax)
    >>> ax_y = fig.add_subplot(gs[1, 1], sharey=ax)
    >>> axis_x = np.arange(0, img.shape[1])
    >>> ax_x.plot(axis_x, img.sum(axis=0))
    >>> ax_x.set_xticks([], [])
    >>> axis_y = np.arange(0, img.shape[0])
    >>> ax_y.plot(img.sum(axis=1), axis_y)
    >>> ax_y.set_yticks([], [])
    >>> plt.show()

    .. plot:: mpl_examples/create_psf_airy.py

    Similarly, we can produce and plot a Gaussian PSF:

    >>> img = create_psf(1*u.um, (40,40), 6*u.um, shape='gauss')

    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from matplotlib.gridspec import GridSpec
    >>> fig = plt.figure(figsize=(6, 6))
    >>> gs = fig.add_gridspec(2, 2,  width_ratios=(4, 1), height_ratios=(1, 4),
    >>>                       left=0.1, right=0.9, bottom=0.1, top=0.9,
    >>>                       wspace=0.05, hspace=0.05)
    >>> ax = fig.add_subplot(gs[1, 0])
    >>> ax.imshow(img)
    >>> ax_x = fig.add_subplot(gs[0, 0], sharex=ax)
    >>> ax_y = fig.add_subplot(gs[1, 1], sharey=ax)
    >>> axis_x = np.arange(0, img.shape[1])
    >>> ax_x.plot(axis_x, img.sum(axis=0))
    >>> ax_x.set_xticks([], [])
    >>> axis_y = np.arange(0, img.shape[0])
    >>> ax_y.plot(img.sum(axis=1), axis_y)
    >>> ax_y.set_yticks([], [])
    >>> plt.show()

    .. plot:: mpl_examples/create_psf_gauss.py

    We can also create a PSF with different F-numbers:

    >>> img = create_psf(1*u.um, (60,40), 6*u.um, shape='gauss')

    >>> import matplotlib.pyplot as plt
    >>> plt.imshow(img, aspect='equal',)
    >>> plt.show()

    .. plot:: mpl_examples/create_psf_gauss_fnum.py


    """

    delta = delta.to(wl.unit)
    if isinstance(fnum, tuple):
        ratio = fnum[1] / fnum[0]
        fnum = fnum[0]
    else:
        ratio = 1

    Nx = int(
        np.round(
            jn_zeros(1, nzero)[-1] / (2.0 * np.pi) * fnum * wl.max() / delta
        ).astype(int)
    )
    Ny = int(
        np.round(
            jn_zeros(1, nzero)[-1]
            / (2.0 * np.pi)
            * fnum
            * ratio
            * wl.max()
            / delta
        ).astype(int)
    )

    if max_array_size is not None:
        max_array_size = list(max_array_size)
        max_array_size[0] = (
            max_array_size[0] // 2
            if max_array_size[0] % 2 == 1
            else (max_array_size[0] + 1) // 2
        )
        max_array_size[1] = (
            max_array_size[1] // 2
            if max_array_size[1] % 2 == 1
            else (max_array_size[1] + 1) // 2
        )

        Nx = max_array_size[0] if Nx > max_array_size[0] else Nx
        Ny = max_array_size[1] if Ny > max_array_size[1] else Ny

    if array_size is not None:
        if array_size[0] == "full":
            if max_array_size is not None:
                Nx = max_array_size[0]
            else:
                logger.error(
                    "max_array_size must be set if array_size is set to full"
                )
                raise ValueError(
                    "max_array_size must be set if array_size is set to full"
                )
        else:
            Nx = (
                array_size[0] // 2
                if array_size[0] % 2 == 1
                else (array_size[0] + 1) // 2
            )

        if array_size[1] == "full":
            if max_array_size is not None:
                Ny = max_array_size[1]
            else:
                logger.error(
                    "max_array_size must be set if array_size is set to full"
                )
                raise ValueError(
                    "max_array_size must be set if array_size is set to full"
                )
        else:
            Ny = (
                array_size[1] // 2
                if array_size[1] % 2 == 1
                else (array_size[1] + 1) // 2
            )

    if shape == "airy":
        d = 1.0 / (fnum * (1.0e-30 * delta.unit + wl))
    elif shape == "gauss":
        sigma = (
            1.029
            * fnum
            * (1.0e-30 * delta.unit + wl)
            / np.sqrt(8.0 * np.log(2.0))
        )
        d = 0.5 / (sigma * sigma)

    x = (
        np.linspace(-Nx * delta.item(), Nx * delta.item(), 2 * Nx + 1)
        * delta.unit
    )
    y = (
        np.linspace(-Ny * delta.item(), Ny * delta.item(), 2 * Ny + 1)
        * delta.unit
    )

    yy, xx = np.meshgrid(y, x)
    yy /= ratio

    if shape == "airy":
        arg = 1.0e-20 * delta.unit + np.pi * np.multiply.outer(
            np.sqrt(yy * yy + xx * xx), d
        )
        arg = arg.value
        img = (j1(arg) / arg) * (j1(arg) / arg)
    elif shape == "gauss":
        arg = np.multiply.outer(yy * yy + xx * xx, d)
        arg = arg.value
        img = np.exp(-arg)

    norm = img.sum(axis=(0, 1))
    img /= norm

    img[..., wl <= 0.0] *= 0.0
    img = np.moveaxis(img, -1, 0)

    return img



import logging
import time

from exosim.log import Logger


class TimedClass(Logger):
    """
    This class adds methods to log the elapsed time
    """

    def __init__(self):
        super().__init__()
        self.start_time_gen = time.time()
        self.start_time = time.time()

    def log_runtime(self, message, level="info"):
        try:
            log_to_call = getattr(self, level)
            time_stamp = time.strftime(
                "%Hh%Mm%Ss", time.gmtime(time.time() - self.start_time)
            )
            full_message = message + ": {}".format(time_stamp)
            log_to_call(full_message)
        except AttributeError:
            self.warning("calling class has no Logger's methods")

        self.start_time = time.time()

    def log_runtime_complete(self, message, level="info"):
        try:
            log_to_call = getattr(self, level)
            time_stamp = time.strftime(
                "%Hh%Mm%Ss", time.gmtime(time.time() - self.start_time_gen)
            )
            full_message = message + ": {}".format(time_stamp)
            log_to_call(full_message)
        except AttributeError:
            self.warning("calling class has no Logger's methods")



import astropy.units as u
import numpy as np
import scipy

from exosim.tasks.task import Task


class ForegroundsToFocalPlane(Task):
    """
    It adds the foreground contribution to the focal plane

    Returns
    -------
     :class:`~exosim.models.signal.Signal`
        focal plane array
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        focal_plane: :class:`~exosim.models.signal.Signal`
            focal plane array (with time evolution)
        paths:  dict
            dictionary containing :class:`~exosim.models.signal.Radiance`
        """

        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("focal_plane", "focal plane")
        self.add_task_param("path", "path")

    def execute(self):
        self.info("adding foreground to focal plane")
        parameters = self.get_task_param("parameters")
        focal_plane = self.get_task_param("focal_plane")
        path = self.get_task_param("path")

        # if check focal plane units
        if focal_plane.data_units == "":
            focal_plane.data_units = u.ct / u.s

        # check if sub focal planes are needed
        if self._look_for_isolate(path):
            self.debug("found isolated contributions")
            sub_focal_plane = focal_plane.copy()
        else:
            sub_focal_plane = None
        sub_focal_planes = {}

        # populate the focal plane with path
        for rad in [k for k in path.keys() if "radiance" in k]:
            if sub_focal_plane:
                tmp_focal_plane = sub_focal_plane.copy()
                tmp_focal_plane.dataset_name = "sub_focal_planes/{}".format(
                    rad
                )

            if "slit_width" in path[rad].metadata.keys():
                # TODO need to perform the convolution on a cube to make it faster
                for t in range(path[rad].time.size):
                    npix = (
                        path[rad].metadata["slit_width"].to(u.um)
                        / parameters["detector"]["delta_pix"]
                        * focal_plane.metadata["oversampling"]
                    )

                    conv = scipy.signal.convolve(
                        path[rad].data[t, 0], np.ones(int(npix)), "same"
                    )
                    focal_plane.data[t, :] += conv
                    if sub_focal_plane:
                        tmp_focal_plane.data[t, :] += conv

            else:
                focal_plane.data[:] += np.nansum(path[rad].data[:, 0])
                if sub_focal_plane:
                    tmp_focal_plane.data[:] += np.nansum(path[rad].data[:, 0])

            if sub_focal_plane:
                sub_focal_planes[rad] = tmp_focal_plane

        self.set_output([focal_plane, sub_focal_planes])

    @staticmethod
    def _look_for_isolate(path):
        rad_list = [rad for rad in path if "radiance" in rad]
        for rad in rad_list:
            if not rad.split("_")[-1].isdigit():
                return True
        return False
